#!/usr/bin/env python3
#
# This file is part of Script of Scripts (sos), a workflow system
# for the execution of commands and scripts in different languages.
# Please visit https://github.com/bpeng2000/SOS
#
# Copyright (C) 2016 Bo Peng (bpeng@mdanderson.org)
##
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

import os
import sys
import re
import subprocess
import tempfile
import shlex
import json
import glob
import copy
import platform
import urllib
import shutil
import zipfile
import gzip
import tarfile
import blessings

from io import BytesIO
from docker import Client
from collections.abc import Sequence
from docker.utils import kwargs_from_env
import multiprocessing as mp
from .utils import env, ProgressBar, natural_keys, transcribe, AbortExecution, short_repr, get_traceback
from .pattern import glob_wildcards
from .sos_eval import interpolate, Undetermined
from .target import FileTarget, fileMD5
from .sos_executor import Base_Executor, MP_Executor, RQ_Executor, Celery_Executor
from .monitor import ProcessMonitor, summarizeExecution

__all__ = ['SoS_Action', 'execute_script', 'sos_run',
    'check_command', 'fail_if', 'warn_if', 'stop_if',
    'download',
    'run', 'bash', 'csh', 'tcsh', 'zsh', 'sh',
    'python', 'python3',
    'perl', 'ruby', 'node', 'JavaScript',
    'R', 'check_R_library',
    'docker_build', 'docker_commit',
    'report', 'pandoc', 'Rmarkdown',
    ]

from .sos_syntax import SOS_RUNTIME_OPTIONS, SOS_ACTION_OPTIONS
from .sos_script import SoS_Script

def get_actions():
    # get the name of all actions, which are identified by an attribute
    # run_mode of the function
    return [k for k, v in globals().items() if hasattr(v, 'run_mode')]

#
# docker support
#
class DockerClient:
    '''A singleton class to ensure there is only one client'''
    _instance = None
    def __new__(cls, *args, **kwargs):
        if not cls._instance:
            cls._instance = super(DockerClient, cls).__new__(cls)
        return cls._instance

    def __init__(self):
        kwargs = kwargs_from_env(assert_hostname=False)
        kwargs.update({'version': 'auto'})
        self.client = Client(**kwargs)
        try:
            self.client.info()
            # mount the /Volumes folder under mac, please refer to
            #    https://github.com/bpeng2000/SOS/wiki/SoS-Docker-guide
            # for details.
            self.has_volumes = False
            if platform.system() == 'Darwin':
                try:
                    # this command log in to the docker machine, check if /Volumes has been mounted,
                    # and try to mount it if possible. This requires users to configure
                    subprocess.call("""docker-machine ssh "{}" 'mount | grep /Volumes || {{ echo "mounting /Volumes"; sudo mount  -t vboxsf Volumes /Volumes; }}' """.format(os.environ['DOCKER_MACHINE_NAME']),
                        shell=True, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
                    env.logger.trace('Sucessfully mount /Volumes to virtual machine')
                    self.has_volumes = True
                except Exception as e:
                    env.logger.trace('Failed to mount /Volumes to virtual machine: {}'.format(e))
        except Exception as e:
            env.logger.debug('Docker client init fail: {}'.format(e))
            self.client = None

    def total_memory(self, image='ubuntu'):
        '''Get the available ram fo the docker machine in Kb'''
        try:
            ret = subprocess.check_output(
                '''docker run -t {} cat /proc/meminfo  | grep MemTotal'''.format(image),
                shell=True, stdin=subprocess.DEVNULL)
            # ret: MemTotal:       30208916 kB
            self.tot_mem = int(ret.split()[1])
        except:
            # some system does not have cat or grep
            self.tot_mem = None
        return self.tot_mem

    def _is_image_avail(self, image):
        images = sum([x['RepoTags'] for x in self.client.images()], [])
        # some earlier version of docker-py returns docker.io/ for global repositories
        images = [x[10:] if x.startswith('docker.io/') else x for x in images]
        return (':' in image and image in images) or \
            (':' not in image and '{}:latest'.format(image) in images)

    def stream(self, line):
        # properly output streamed output
        try:
            sys.stdout.write(json.loads(line).get('stream', ''))
        except ValueError:
            # sometimes all the data is sent on a single line ????
            #
            # ValueError: Extra data: line 1 column 87 - line 1 column
            # 33268 (char 86 - 33267)
            # This ONLY works because every line is formatted as
            # {"stream": STRING}
            for obj in re.findall('{\s*"stream"\s*:\s*"[^"]*"\s*}', line):
                sys.stdout.write(json.loads(obj).get('stream', ''))

    def build(self, script, **kwargs):
        if not self.client:
            raise RuntimeError('Cannot connect to the Docker daemon. Is the docker daemon running on this host?')
        if script is not None:
            f = BytesIO(script.encode('utf-8'))
            for line in self.client.build(fileobj=f, **kwargs):
                self.stream(line.decode())
        else:
            for line in self.client.build(**kwargs):
                self.stream(line.decode())
        # if a tag is given, check if the image is built
        if 'tag' in kwargs and not self._is_image_avail(kwargs['tag']):
            raise RuntimeError('Image with tag {} is not created.'.format(kwargs['tag']))

    def import_image(self, image, **kwargs):
        if not self.client:
            raise RuntimeError('Cannot connect to the Docker daemon. Is the docker daemon running on this host?')
        env.logger.info('docker import {}'.format(image))
        self.client.import_image(image, **kwargs)

    def pull(self, image):
        if not self.client:
            raise RuntimeError('Cannot connect to the Docker daemon. Is the docker daemon running on this host?')
        # if image is specified, check if it is available locally. If not, pull it
        ret = 0
        if not self._is_image_avail(image):
            env.logger.info('docker pull {}'.format(image))
            # using subprocess instead of docker-py's pull function because this would have
            # much better progress bar display
            ret = subprocess.call('docker pull {}'.format(image), shell=True)
            #for line in self.client.pull(image, stream=True):
            #    self.stream(line)
        if not self._is_image_avail(image):
            raise RuntimeError('Failed to pull image {}'.format(image))
        return ret

    def commit(self, **kwargs):
        if not self.client:
            raise RuntimeError('Cannot connect to the Docker daemon. Is the docker daemon running on this host?')
        for line in self.client.commit(**kwargs):
            self.stream(line.decode())
        return 0

    def run(self, image, script='', interpreter='', suffix='.sh', **kwargs):
        if self.client is None:
            raise RuntimeError('Cannot connect to the Docker daemon. Is the docker daemon running on this host?')
        #
        env.logger.debug('docker_run with keyword args {}'.format(kwargs))
        #
        # now, write a temporary file to a tempoary directory under the current directory, this is because
        # we need to share the directory to ...
        with tempfile.TemporaryDirectory(dir=os.getcwd()) as tempdir:
            # keep the temporary script for debugging purposes
            # tempdir = tempfile.mkdtemp(dir=os.getcwd())
            if script:
                tempscript = 'docker_run_{}{}'.format(os.getpid(), suffix)
                with open(os.path.join(tempdir, tempscript), 'w') as script_file:
                    script_file.write(script)
            #
            binds = []
            if 'volumes' in kwargs:
                volumes = [kwargs['volumes']] if isinstance(kwargs['volumes'], str) else kwargs['volumes']
                for vol in volumes:
                    if not vol:
                        continue
                    if vol.count(':') != 1:
                        raise RuntimeError('Please specify columes in the format of host_dir:mnt_dir')
                    host_dir, mnt_dir = vol.split(':')
                    if platform.system() == 'Darwin':
                        # under Darwin, host_dir must be under /Users
                        if not os.path.abspath(host_dir).startswith('/Users') and not (self.has_volumes and os.path.abspath(host_dir).startswith('/Volumes')):
                            raise RuntimeError('hostdir ({}) under MacOSX must be under /Users or /Volumes (if properly configured, see https://github.com/bpeng2000/SOS/wiki/SoS-Docker-guide for details) to be usable in docker container'.format(host_dir))
                    binds.append('{}:{}'.format(os.path.abspath(host_dir), mnt_dir))
            #
            volumes_opt = ' '.join('-v {}'.format(x) for x in binds)
            # under mac, we by default share /Users within docker
            if platform.system() == 'Darwin':
                if not any(x.startswith('/Users:') for x in binds):
                    volumes_opt += ' -v /Users:/Users'
                if self.has_volumes:
                    volumes_opt += ' -v /Volumes:/Volumes'
            if not any(x.startswith('/tmp:') for x in binds):
                volumes_opt += ' -v /tmp:/tmp'
            #
            mem_limit_opt = ''
            if 'mem_limit' in kwargs:
                mem_limit_opt = '--memory={}'.format(kwargs['mem_limit'])
            #
            volumes_from_opt = ''
            if 'volumes_from' in kwargs:
                if isinstance(kwargs['volumes_from'], str):
                    volumes_from_opt = '--volumes_from={}'.format(kwargs['volumes_from'])
                elif isinstance(kwargs['volumes_from'], list):
                    volumes_from_opt = ' '.join('--volumes_from={}'.format(x) for x in kwargs['volumes_from'])
                else:
                    raise RuntimeError('Option volumes_from only accept a string or list of string'.format(kwargs['volumes_from']))
            # we also need to mount the script
            cmd_opt = ''
            if script and interpreter:
                volumes_opt += ' -v {}:{}'.format(os.path.join(tempdir, tempscript), '/var/lib/sos/{}'.format(tempscript))
                cmd_opt = interpreter.replace('{}', '/var/lib/sos/{}'.format(tempscript))
            #
            working_dir_opt = '-w={}'.format(os.path.abspath(os.getcwd()))
            if 'working_dir' in kwargs:
                if not os.path.isabs(kwargs['working_dir']):
                    env.logger.warning('An absolute path is needed for -w option of docker run command. "{}" provided, "{}" used.'
                        .format(kwargs['working_dir'], os.path.abspath(os.path.expanduser(kwargs['working_dir']))))
                    working_dir_opt = '-w={}'.format(os.path.abspath(os.path.expanduser(kwargs['working_dir'])))
                else:
                    working_dir_opt = '-w={}'.format(kwargs['working_dir'])
            #
            env_opt = ''
            if 'environment' in kwargs:
                if isinstance(kwargs['environment'], dict):
                    env_opt = ' '.join('-e {}={}'.format(x,y) for x,y in kwargs['environment'].items())
                elif isinstance(kwargs['environment'], list):
                    env_opt = ' '.join('-e {}'.format(x) for x in kwargs['environment'])
                elif isinstance(kwargs['environment'], str):
                    env_opt = '-e {}'.format(kwargs['environment'])
                else:
                    raise RuntimeError('Invalid value for option environment (str, list, or dict is allowd, {} provided)'.format(kwargs['environment']))
            #
            port_opt = '-P'
            if 'port' in kwargs:
                if isinstance(kwargs['port'], (str, int)):
                    port_opt = '-p {}'.format(kwargs['port'])
                elif isinstance(kwargs['port'], list):
                    port_opt = ' '.join('-p {}'.format(x) for x in kwargs['port'])
                else:
                    raise RuntimeError('Invalid value for option port (a list of intergers), {} provided'.format(kwargs['port']))
            #
            name_opt = ''
            if 'name' in kwargs:
                name_opt = '--name={}'.format(kwargs['name'])
            #
            stdin_opt = ''
            if 'stdin_open' in kwargs and kwargs['stdin_optn']:
                stdin_opt = '-i'
            #
            tty_opt = '-t'
            if 'tty' in kwargs and not kwargs['tty']:
                tty_opt = ''
            #
            user_opt = ''
            if 'user' in kwargs:
                user_opt = '-u {}'.format(kwargs['user'])
            #
            extra_opt = ''
            if 'extra_args' in kwargs:
                extra_opt = kwargs['extra_args']
            #
            security_opt = ''
            if platform.system() == 'Linux':
                # this is for a selinux problem when /var/sos/script cannot be executed
                security_opt = '--security-opt label:disable'
            command = 'docker run --rm {} {} {} {} {} {} {} {} {} {} {} {} {} {}'.format(
                security_opt,       # security option
                volumes_opt,        # volumes
                volumes_from_opt,   # volumes_from
                name_opt,           # name
                stdin_opt,          # stdin_optn
                tty_opt,            # tty
                port_opt,           # port
                working_dir_opt,    # working dir
                user_opt,           # user
                env_opt,            # environment
                mem_limit_opt,      # memory limit
                extra_opt,          # any extra parameters
                image,              # image
                cmd_opt
                )
            env.logger.info(command)
            ret = subprocess.call(command, shell=True)
            if ret != 0:
                msg = 'The script has been saved to .sos/{} so that you can execute it using the following command:\n{}'.format(
                    tempscript, command.replace(tempdir, os.path.abspath('./.sos')))
                shutil.copy(os.path.join(tempdir, tempscript), '.sos')
                if ret == 125:
                    raise RuntimeError('Docker daemon failed (exitcode=125). ' + msg)
                elif ret == 126:
                    raise RuntimeError('Failed to invoke specified command (exitcode=126). ' + msg)
                elif ret == 127:
                    raise RuntimeError('Failed to locate specified command (exitcode=127). ' + msg)
                elif ret == 137:
                    if not hasattr(self, 'tot_mem'):
                        self.tot_mem = self.total_memory(image)
                    if self.tot_mem is None:
                        raise RuntimeError('Script killed by docker. ' + msg)
                    else:
                        raise RuntimeError('Script killed by docker, probably because of lack of RAM (available RAM={:.1f}GB, exitcode=137). '.format(self.tot_mem/1024/1024) + msg)
                else:
                    raise RuntimeError('Executing script in docker returns an error (exitcode={}). '.format(ret) + msg)
        return 0

#
# A decoration function that allows SoS to replace all SoS actions
# with a null action.
#
def SoS_Action(run_mode=['run', 'interactive']):
    run_mode = [run_mode] if isinstance(run_mode, str) else run_mode
    def runtime_decorator(func):
        def action_wrapper(*args, **kwargs):
            try:
                old_mode = env.run_mode
                if 'run_mode' in kwargs:
                    if not isinstance(kwargs['run_mode'], str):
                        raise ValueError('A string is required for option run_mode, {} provided.'
                            .format(kwargs['run_mode']))
                    if kwargs['run_mode'] not in ['prepare', 'interactive']:
                        raise ValueError('Parameter run_mode can only be prepare, or interactive')
                    top_mode = 'run' if 'run' in run_mode else 'prepare'
                    if env.run_mode == top_mode and top_mode != kwargs['run_mode']:
                        # if run has been moved away, do nothing
                        return None
                    elif env.run_mode == kwargs['run_mode']:
                        env.run_mode = top_mode

                # docker files will be downloaded in run or prepare mode
                if 'docker_file' in kwargs and env.run_mode in ['prepare', 'run', 'interactive']:
                    docker = DockerClient()
                    docker.import_image(kwargs['docker_file'])
                # handle image
                if 'docker_image' in kwargs:
                    docker = DockerClient()
                    docker.pull(kwargs['docker_image'])
                    if env.run_mode in ['prepare', 'interactive']:
                        mem = docker.total_memory(kwargs['docker_image'])
                        if mem is not None:
                            if mem < 4000000: # < 4G
                                env.logger.warning('Docker machine has {:.1f} GB of total memory and might not be enough for your operation. Please refer to https://github.com/bpeng2000/SOS/wiki/SoS-Docker-guide to adjust the docker machine if needed.'
                                    .format(mem/1024/1024))
                            else:
                                env.logger.debug('Docker machine has {:.1f} GB of total memory ram'.format(mem/1024/1024))
                if env.run_mode not in run_mode:
                    # return dynamic expression when not in run mode, that is to say
                    # the script logic cannot rely on the result of the action
                    return Undetermined(func.__name__)
                if env.run_mode == 'interactive':
                    for k,v in kwargs.items():
                        if k in SOS_RUNTIME_OPTIONS and k not in SOS_ACTION_OPTIONS:
                            env.logger.warning('Passing runtime option "{0}" to action is deprecated. Please use "task: {0}={1}" before action instead.'.format(k, v))
                if 'active' in kwargs:
                    if isinstance(kwargs['active'], int):
                        if kwargs['active'] >= 0 and env.sos_dict['_index'] != kwargs['active']:
                            return None
                        if kwargs['active'] < 0 and env.sos_dict['_index'] != kwargs['active'] + env.sos_dict['__num_groups__']:
                            return None
                    elif isinstance(kwargs['active'], Sequence):
                        allowed_index = list([x if x >= 0 else env.sos_dict['__num_groups__'] + x for x in kwargs['active']])
                        if env.sos_dict['_index'] not in allowed_index:
                            return None
                    elif isinstance(kwargs['active'], slice):
                        allowed_index = list(range(env.sos_dict['__num_groups__']))[kwargs['active']]
                        if env.sos_dict['_index'] not in allowed_index:
                            return None
                    else:
                        raise RuntimeError('Unacceptable value for option active: {}'.format(kwargs['active']))
                if 'workdir' in kwargs:
                    if not kwargs['workdir'] or not isinstance(kwargs['workdir'], str):
                        raise RuntimeError('workdir option should be a path, {} provided'.format(kwargs['workdir']))
                    if not os.path.isdir(os.path.expanduser(kwargs['workdir'])):
                        os.makedirs(os.path.expanduser(kwargs['workdir']))
                    try:
                        olddir = os.getcwd()
                        os.chdir(os.path.expanduser(kwargs['workdir']))
                        res = func(*args, **kwargs)
                    finally:
                        os.chdir(olddir)
                else:
                    res = func(*args, **kwargs)
                return res
            finally:
                # restore run_mode is the run mode is changed by parameter run_mode
                env.run_mode = old_mode
        action_wrapper.run_mode = run_mode
        return action_wrapper
    return runtime_decorator


class SoS_ExecuteScript:
    def __init__(self, script, interpreter, suffix, args=''):
        self.script = script
        self.interpreter = interpreter
        if args:
            self.interpreter += ' ' + args
        self.suffix = suffix

    def run(self, **kwargs):
        transcribe(self.script, action=self.interpreter)
        if env.run_mode == 'prepare':
            check_command(self.interpreter.split()[0], quiet=True)
            return
        if '{}' not in self.interpreter:
            self.interpreter += ' {}'
        debug_script_file = os.path.join(env.exec_dir, '.sos', '{}_{}{}'.format(env.sos_dict['step_name'],
            env.sos_dict['_index'], self.suffix))
        env.logger.debug('Script for step {} is saved to {}'.format(env.sos_dict['step_name'], debug_script_file))
        with open(debug_script_file, 'w') as sfile:
            sfile.write(self.script)
        if 'docker_image' in kwargs:
            docker = DockerClient()
            docker.run(kwargs['docker_image'], self.script, self.interpreter, self.suffix,
                **kwargs)
        else:
            try:
                script_file = tempfile.NamedTemporaryFile(mode='w+t', suffix=self.suffix, delete=False).name
                with open(script_file, 'w') as sfile:
                    sfile.write(self.script)
                cmd = self.interpreter.replace('{}', shlex.quote(script_file))
                #
                if env.run_mode == 'interactive':
                    # need to catch output and send to python output, which will in trun be hijacked by SoS notebook
                    p = subprocess.Popen(cmd, shell=True, stderr=subprocess.PIPE, stdout=subprocess.PIPE)
                    pid = p.pid
                    m = ProcessMonitor(pid, msg=self.script)
                    m.start()
                    env.register_process(p.pid, 'Runing {}'.format(script_file))
                    out, err = p.communicate()
                    sys.stdout.write(out.decode())
                    sys.stderr.write(err.decode())
                    ret = p.returncode
                    summarizeExecution(pid)
                    sys.stdout.flush()
                    sys.stderr.flush()
                else:
                    p = subprocess.Popen(cmd, shell=True)
                    pid = p.pid
                    m = ProcessMonitor(pid, msg=self.script)
                    m.start()
                    env.register_process(pid, 'Runing {}'.format(script_file))
                    ret = p.wait()
                    summarizeExecution(pid)

                if ret != 0:
                    with open(debug_script_file, 'w') as sfile:
                        sfile.write(self.script)
                    cmd = self.interpreter.replace('{}', shlex.quote(debug_script_file))
                    raise RuntimeError('Failed to execute script. The script is saved to {}. Please use command "{}" under {} to test it.'
                        .format(debug_script_file, cmd, os.getcwd()))
            except Exception as e:
                env.logger.error('Failed to execute script: {}'.format(e))
                raise
            finally:
                env.deregister_process(p.pid)
                os.remove(script_file)


@SoS_Action(run_mode=['prepare', 'run', 'interactive'])
def sos_run(workflow, **kwargs):
    '''Execute a workflow from specified source, input, and output
    By default the workflow is defined in the existing SoS script, but
    extra sos files can be specified from paramter source. The workflow
    will be execute in the current step namespace with _input as workflow
    input. '''
    script = SoS_Script(env.sos_dict['__step_context__'].content, env.sos_dict['__step_context__'].filename)
    wf = script.workflow(workflow)
    # if wf contains the current step or one of the previous one, this constitute
    # recusive nested workflow and should not be allowed
    if env.sos_dict['step_name'] in ['{}_{}'.format(x.name, x.index) for x in wf.sections]:
        raise RuntimeError('Nested workflow {} contains the current step {}'.format(workflow, env.sos_dict['step_name']))
    # for nested workflow, _input would becomes the input of workflow.
    for k,v in kwargs.items():
        env.sos_dict.set(k, v)
    env.sos_dict.set('__step_output__', copy.deepcopy(env.sos_dict['_input']))
    if env.run_mode == 'dryrun':
        env.logger.info('Checking nested workflow {}'.format(workflow))
        return Base_Executor(wf, args=env.sos_dict['__args__'], nested=True).dryrun()
    elif env.run_mode == 'prepare':
        env.logger.info('Preparing nested workflow {}'.format(workflow))
        return Base_Executor(wf, args=env.sos_dict['__args__'], nested=True).prepare()
    elif env.run_mode == 'run':
        env.logger.info('Executing workflow ``{}`` with input ``{}``'
            .format(workflow, short_repr(env.sos_dict['_input'], True)))
        #
        # NOTE: Because at run mode we do not really have the DAG (which might have
        # been changed and need to be re-prepared) so it is necessary to prepare
        # the workflow at run mode.
        #
        if env.__task_engine__ is None:
            if env.max_jobs == 1:
                return Base_Executor(wf, args=env.sos_dict['__args__'], nested=True).run()
            else:
                return MP_Executor(wf, args=env.sos_dict['__args__'], nested=True).run()
        elif env.__task_engine__ == 'rq':
            return RQ_Executor(wf, args=env.sos_dict['__args__'], nested=True).run()
        elif env.__task_engine__ == 'celery':
            return Celery_Executor(wf, args=env.sos_dict['__args__'], nested=True).run()
    elif env.run_mode == 'interactive':
        raise RuntimeError('Action sos_run is not supported in interactive mode')

@SoS_Action(run_mode=['prepare', 'run', 'interactive'])
def execute_script(script, interpreter, suffix, args='', **kwargs):
    return SoS_ExecuteScript(script, interpreter, suffix, args).run(**kwargs)

_check_command_cache = {}
@SoS_Action(run_mode=['prepare', 'run', 'interactive'])
def check_command(cmd, pattern = None, quiet=False):
    '''Raise an exception if output of `cmd` does not match specified `pattern`.
    Multiple patterns can be specified as a list of patterns.
    When pattern is None, check the existence of command `cmd`
    and raise an error if command does not exist.'''
    global _check_command_cache
    if cmd in _check_command_cache:
        return _check_command_cache[cmd]
    ret_val = 0
    cmd_name = shlex.split(cmd)[0]
    full_name = shutil.which(cmd_name)
    if not full_name:
        raise ValueError('Command ``{}`` not found!'.format(cmd_name))
    if not quiet:
        env.logger.info('Command ``{}`` is located as ``{}``.'.format(cmd, full_name))
    if pattern is None and len(shlex.split(cmd)) == 1:
        return 0
    try:
        output = subprocess.check_output(cmd, stderr=subprocess.STDOUT, shell=True, timeout=2).decode()
    except subprocess.TimeoutExpired as e:
        output = e.output.decode()
        ret_val = 1
        env.logger.warning(e)
        #env.logger.warning(e.output.decode())
    except subprocess.CalledProcessError as e:
        ret_val = e.returncode
        output = e.output.decode()
        env.logger.warning(e)
    #
    env.logger.trace('Output of command ``{}`` is ``{}``'.format(cmd, output))
    #
    if pattern:
        pattern = [pattern] if isinstance(pattern, str) else pattern
        if all([re.search(x, output, re.MULTILINE) is None for x in pattern]):
            raise ValueError('Output of command ``{}`` does not match specified regular expression ``{}``.'
                .format(cmd, ' or '.join(pattern)))
    _check_command_cache[cmd] = ret_val
    return ret_val

@SoS_Action(run_mode=['prepare', 'run', 'interactive'])
def fail_if(expr, msg=''):
    '''Raise an exception with `msg` if condition `expr` is False'''
    if expr:
        raise RuntimeError(msg)
    return 0

@SoS_Action(run_mode=['prepare', 'run', 'interactive'])
def warn_if(expr, msg=''):
    '''Yield an warning message `msg` if `expr` is False '''
    if expr:
        env.logger.warning(msg)
    return 0

@SoS_Action(run_mode=['prepare', 'run', 'interactive'])
def stop_if(expr, msg=''):
    '''Abort the execution of the current step or loop and yield
    an warning message `msg` if `expr` is False '''
    if expr:
        raise AbortExecution(msg)
    return 0

#
# download file with progress bar
#
def downloadURL(URL, dest, decompress=False, index=None):
    dest = os.path.abspath(os.path.expanduser(dest))
    dest_dir, filename = os.path.split(dest)
    #
    if not os.path.isdir(dest_dir):
        os.makedirs(dest_dir)
    if not os.path.isdir(dest_dir):
        raise RuntimeError('Failed to create destination directory to download {}'.format(URL))
    #
    message = filename
    if len(message) > 30:
        message = message[:10] + '...' + message[-16:]
    #
    dest_tmp = dest + '.tmp_{}'.format(os.getpid())
    term_width = shutil.get_terminal_size((80, 20)).columns
    try:
        env.logger.debug('Download {} to {}'.format(URL, dest))
        prog = ProgressBar(message, disp=env.verbosity > 1)
        sig = FileTarget(dest)
        if os.path.isfile(dest):
            if env.sig_mode == 'construct':
                prog.done(message + ': \033[32m writing signature\033[0m')
                sig.write_sig()
                prog.done(message + ': \033[32m signature calculated\033[0m')
                return True
            elif env.sig_mode == 'ignore':
                prog.done(message + ': \033[32m use existing\033[0m')
                return True
            else:
                prog.done(message + ': \033[32m Validating signature\033[0m')
                if sig.validate():
                    prog.done(message + ': \033[32m Validated\033[0m')
                    return True
                else:
                    prog.done(message + ':\033[91m Signature mismatch\033[0m')
        #
        # Stop using pycurl because of libcurl version compatibility problems
        # that happen so often and difficult to fix. Error message looks like
        #
        # Reason: Incompatible library version: pycurl.cpython-35m-darwin.so
        # requires version 9.0.0 or later, but libcurl.4.dylib provides version 7.0.0
        #
        #with open(dest_tmp, 'wb') as f:
        #    c = pycurl.Curl()
        #    c.setopt(pycurl.URL, str(URL))
        #    c.setopt(pycurl.WRITEFUNCTION, f.write)
        #    c.setopt(pycurl.SSL_VERIFYPEER, False)
        #    c.setopt(pycurl.NOPROGRESS, False)
        #    c.setopt(pycurl.PROGRESSFUNCTION, prog.curlUpdate)
        #    c.perform()
        #if c.getinfo(pycurl.HTTP_CODE) == 404:
        #    prog.done(message + ':\033[91m 404 Error {}\033[0m'.format(' '*(term_width - len(message) - 12)))
        #    try:
        #        os.remove(dest_tmp)
        #    except OSError:
        #        pass
        #    return False
        with open(dest_tmp, 'wb') as f:
            try:
                u = urllib.request.urlopen(str(URL))
                try:
                    file_size = int(u.getheader("Content-Length"))
                except:
                    file_size = None
                file_size_dl = 0
                block_sz = 8192
                while True:
                    buffer = u.read(block_sz)
                    if not buffer:
                        break
                    file_size_dl += len(buffer)
                    f.write(buffer)
                    prog.urllibUpdate(file_size, file_size_dl)
            except urllib.error.HTTPError as e:
                prog.done(message + ':\033[91m {} Error\033[0m'.format(e.code))
                try:
                    os.remove(dest_tmp)
                except OSError:
                    pass
                return False
            except Exception as e:
                prog.done(message + ':\033[91m {}\033[0m'.format(e))
                try:
                    os.remove(dest_tmp)
                except OSError:
                    pass
                return False
        #
        os.rename(dest_tmp, dest)
        decompressed = 0
        if decompress:
            if zipfile.is_zipfile(dest):
                prog.done(message + ':\033[91m Decompressing\033[0m')
                zip = zipfile.ZipFile(dest)
                zip.extractall(dest_dir)
                names = zip.namelist()
                for name in names:
                    if not os.path.isfile(os.path.join(dest_dir, name)):
                        return False
                    else:
                        sig.add(os.path.join(dest_dir, name))
                        decompressed += 1
            elif tarfile.is_tarfile(dest):
                prog.done(message + ':\033[91m Decompressing\033[0m')
                with tarfile.open(dest, 'r:*') as tar:
                    tar.extractall(dest_dir)
                    # only extract files
                    files = [x.name for x in tar.getmembers() if x.isfile()]
                    for name in files:
                        if not os.path.isfile(os.path.join(dest_dir, name)):
                            return False
                        else:
                            sig.add(os.path.join(dest_dir, name))
                            decompressed += 1
            elif dest.endswith('.gz'):
                prog.done(message + ':\033[91m Decompressing\033[0m')
                decomp = dest[:-3]
                with gzip.open(dest, 'rb') as fin, open(decomp, 'wb') as fout:
                    buffer = fin.read(100000)
                    while buffer:
                        fout.write(buffer)
                        buffer = fin.read(100000)
                sig.add(decomp)
                decompressed += 1
        decompress_msg = '' if not decompressed else ' ({} file{} decompressed)'.format(
            decompressed, '' if decompressed <= 1 else 's')
        prog.done(message + ':\033[32m downloaded{} {}\033[0m'.format(decompress_msg,
            ' '*(term_width - len(message) - 13 - len(decompress_msg))))
        # if a md5 file exists
        # if downloaded files contains .md5 signature, use them to validate
        # downloaded files.
        if os.path.isfile(dest + '.md5'):
            prog.done(message + ':\033[91m Verifying md5 signature\033[0m')
            with open(dest + '.md5') as md5:
                rec_md5 = md5.readline().split()[0].strip()
                obs_md5 = fileMD5(dest, partial=False)
                if rec_md5 != obs_md5:
                    prog.done(message + ':\033[91m MD5 signature mismatch\033[0m')
                    env.logger.warning('md5 signature mismatch for downloaded file {} (recorded {}, observed {})'
                        .format(filename[:-4], rec_md5, obs_md5))
            prog.done(message + ':\033[91m MD5 signature verified\033[0m')
    except Exception as e:
        if env.verbosity > 2:
             sys.stderr.write(get_traceback())
        env.logger.error('Failed to download: {}'.format(e))
        return False
    finally:
        # if there is something wrong still remove temporary file
        if os.path.isfile(dest_tmp):
            os.remove(dest_tmp)
    sig.write_sig()
    return os.path.isfile(dest)


@SoS_Action(run_mode=['prepare', 'run', 'interactive'])
def download(URLs, dest_dir='.', dest_file=None, decompress=False):
    '''Download files from specified URL, which should be space, tab or
    newline separated URLs. The files will be downloaded to specified
    destination. If `filename.md5` files are downloaded, they are used to
    validate downloaded `filename`. Unless otherwise specified, compressed
    files are decompressed.
    '''
    if isinstance(URLs, str):
        urls = [x.strip() for x in URLs.split() if x.strip()]
    else:
        urls = list(URLs)
    #
    if dest_file is not None and len(urls) != 1:
        raise RuntimeError('Only one URL is allowed if a destination file is specified.')
    #
    if dest_file is None:
        filenames = []
        for idx, url in enumerate(urls):
            token = urllib.parse.urlparse(url)
            # if no scheme or netloc, the URL is not acceptable
            if not all([getattr(token, qualifying_attr) for qualifying_attr in  ('scheme', 'netloc')]):
                filenames.append(None)
                continue
            filename = os.path.split(token.path)[-1]
            if not filename:
                filenames.append(None)
                continue
            filenames.append(os.path.join(dest_dir, filename))
    else:
        filenames = [dest_file]
    #
    succ = [False for x in urls]
    if len(succ) > 1:
        # first scroll several lines to reserve place for progress bar
        for url in urls:
            sys.stderr.write('\n')
        with mp.Pool(processes = env.sos_dict['CONFIG'].get('sos_download_processes', 5)) as pool:
            for idx, (url, filename) in enumerate(zip(urls, filenames)):
                if not filename:
                    continue
                succ[idx] = pool.apply_async(downloadURL, (url, filename,
                    decompress, len(urls) - idx))
            succ = [x.get() if isinstance(x, mp.pool.AsyncResult) else x for x in succ]
        #
        if sys.platform != 'win32':
            t = blessings.Terminal(stream=sys.stderr)
            sys.stderr.write(t.move( t.height, 0)) # + '\n')
    else:
        if dest_file is not None:
            succ[0] = downloadURL(urls[0], dest_file, decompress=decompress)
        else:
           if filenames[0]:
                succ[0] = downloadURL(urls[0], filenames[0], decompress=decompress)
    #
    #for su, url in zip(succ, urls):
    #    if not su:
    #        env.logger.warning('Failed to download {}'.format(url))
    if not all(succ):
        raise RuntimeError('Not all files have been downloaded')
    return 0

@SoS_Action(run_mode=['prepare', 'run', 'interactive'])
def run(script, args='', **kwargs):
    return SoS_ExecuteScript(script, '/bin/bash', '.sh', args).run(**kwargs)

@SoS_Action(run_mode=['prepare', 'run', 'interactive'])
def bash(script, args='', **kwargs):
    return SoS_ExecuteScript(script, '/bin/bash', '.sh', args).run(**kwargs)

@SoS_Action(run_mode=['prepare', 'run', 'interactive'])
def csh(script, args='', **kwargs):
    return SoS_ExecuteScript(script, '/bin/csh', '.csh', args).run(**kwargs)

@SoS_Action(run_mode=['prepare', 'run', 'interactive'])
def tcsh(script, args='', **kwargs):
    return SoS_ExecuteScript(script, '/bin/tcsh', '.sh', args).run(**kwargs)

@SoS_Action(run_mode=['prepare', 'run', 'interactive'])
def zsh(script, args='', **kwargs):
    return SoS_ExecuteScript(script, '/bin/zsh', '.zsh', args).run(**kwargs)

@SoS_Action(run_mode=['prepare', 'run', 'interactive'])
def sh(script, args='', **kwargs):
    return SoS_ExecuteScript(script, '/bin/sh', '.sh', args).run(**kwargs)

@SoS_Action(run_mode=['prepare', 'run', 'interactive'])
def python(script, args='', **kwargs):
    return SoS_ExecuteScript(script, 'python', '.py', args).run(**kwargs)

@SoS_Action(run_mode=['prepare', 'run', 'interactive'])
def python3(script, args='', **kwargs):
    return SoS_ExecuteScript(script, 'python3', '.py', args).run(**kwargs)

@SoS_Action(run_mode=['prepare', 'run', 'interactive'])
def perl(script, args='', **kwargs):
    return SoS_ExecuteScript(script, 'perl', '.pl', args).run(**kwargs)

@SoS_Action(run_mode=['prepare', 'run', 'interactive'])
def ruby(script, args='', **kwargs):
    return SoS_ExecuteScript(script, 'ruby', '.rb', args).run(**kwargs)

@SoS_Action(run_mode=['prepare', 'run', 'interactive'])
def node(script, args='', **kwargs):
    return SoS_ExecuteScript(script, 'node', '.js', args).run(**kwargs)

@SoS_Action(run_mode=['prepare', 'run', 'interactive'])
def JavaScript(script, args='', **kwargs):
    return SoS_ExecuteScript(script, 'node', '.js', args).run(**kwargs)

@SoS_Action(run_mode=['prepare', 'run', 'interactive'])
def R(script, args='', **kwargs):
    # > getOption('defaultPackages')
    # [1] "datasets"  "utils"     "grDevices" "graphics"  "stats"     "methods"
    return SoS_ExecuteScript(
        script, 'Rscript --default-packages=datasets,methods,utils,stats,grDevices,graphics ', '.R', args).run(**kwargs)

@SoS_Action(run_mode=['prepare', 'run', 'interactive'])
def check_R_library(name, version = None, repos = 'http://cran.us.r-project.org'):
    '''Check existence and version match of R library.
    cran and bioc packages are unique yet might overlap with github.
    Therefore if the input name is {repo}/{pkg} the package will be
    installed from github if not available, else from cran or bioc
    '''
    check_command('Rscript', quiet=True)
    output_file = tempfile.NamedTemporaryFile(mode='w+t', suffix='.txt', delete=False).name
    if len(glob_wildcards('{repo}/{pkg}', [name])['repo']):
        # package is from github
        check_R_library('devtools')
        install_script = interpolate('''
        options(warn=-1)
        package_repo <- ${name!r}
        package <- basename(package_repo)
        if (require(package, character.only=TRUE, quietly=TRUE)) {
            write(paste(package, packageVersion(package), "AVAILABLE"), file="${output_file}")
        } else {
            devtools::install_github(package_repo)
            # if it still does not exist, write the package name to output
            if (require(package, character.only=TRUE, quietly=TRUE)) {
                write(paste(package, packageVersion(package), "INSTALLED"), file="${output_file}")
            } else {
                write(paste(package, "NA", "MISSING"), file="${output_file}")
                quit("no")
            }
        }
        cur_version <- packageVersion(package)
        ''', '${ }', locals())
    else:
        # package is from cran or bioc
        install_script = interpolate('''
        options(warn=-1)
        package <- ${name!r}
        if (require(package, character.only=TRUE, quietly=TRUE)) {
            write(paste(package, packageVersion(package), "AVAILABLE"), file="${output_file}")
        } else {
            install.packages(package, repos="${repos}",
                quiet=FALSE)
            # if the package still does not exist
            if (!require(package, character.only=TRUE, quietly=TRUE)) {
                source("http://bioconductor.org/biocLite.R")
                biocLite(package, suppressUpdates=TRUE, suppressAutoUpdate=TRUE, ask=FALSE)
            }
            # if it still does not exist, write the package name to output
            if (require(package, character.only=TRUE, quietly=TRUE)) {
                write(paste(package, packageVersion(package), "INSTALLED"), file="${output_file}")
            } else {
                write(paste(package, "NA", "MISSING"), file="${output_file}")
                quit("no")
            }
        }
        cur_version <- packageVersion(package)
        ''', '${ }', locals())
    version_script = ''
    if version is not None:
        version = [version] if isinstance(version, str) else version
        operators = []
        for idx, value in enumerate(version):
            value = str(value)
            if value.endswith('+'):
                operators.append('>=')
                version[idx] = value[:-1]
            elif value.endswith('-'):
                operators.append('<')
                version[idx] = value[:-1]
            else:
                operators.append('==')
        # check version and mark version mismatch
        # if current version satisfies any of the
        # requirement the check program quits
        for x, y in zip(version, operators):
            version_script += '''
            if (cur_version {1} {0}) {{
              quit("no")
            }}
            '''.format(repr(x), y)
        version_script += 'write(paste(package, cur_version, "VERSION_MISMATCH"), file = {})'.\
          format(repr(output_file))
    # temporarily change the run mode to run to execute script
    try:
        env.run_mode = 'run'
        SoS_ExecuteScript(install_script + version_script, 'Rscript --default-packages=utils', '.R').run()
    finally:
        env.run_mode = 'prepare'
    ret_val = 0
    with open(output_file) as tmp:
        for line in tmp:
            lib, version, status = line.split()
            if status.strip() == "MISSING":
                raise RuntimeError('R Library {} is not available and cannot be installed.'.format(lib))
            elif status.strip() == 'AVAILABLE':
                env.logger.info('R library {} ({}) is available'.format(lib, version))
            elif status.strip() == 'INSTALLED':
                env.logger.info('R library {} ({}) has been installed'.format(lib, version))
            elif status.strip() == 'VERSION_MISMATCH':
                env.logger.warning('R library {} ({}) does not satisfy version requirement!'.format(lib, version))
                ret_val = 1
            else:
                raise RuntimeError('This should not happen: {}'.format(line))
    try:
        os.remove(output_file)
    except:
        pass
    return ret_val


@SoS_Action(run_mode=['run', 'interactive'])
def docker_build(dockerfile=None, **kwargs):
    '''docker build command. By default a script is sent to the docker build command but
    you can also specify different parameters defined inu//docker-py.readthedocs.org/en/stable/api/#build
    '''
    docker = DockerClient()
    docker.build(dockerfile, **kwargs)
    return 0


@SoS_Action(run_mode=['run', 'interactive'])
def docker_commit(**kwargs):
    docker = DockerClient()
    docker.commit(**kwargs)
    return 0

@SoS_Action(run_mode=['run', 'interactive'])
def report(script=None, from_file=None, to_file=None, mode='a', **kwargs):
    if to_file is not None:
        if not to_file:
            raise RuntimeError('Invalid parameter to_file "{}"'.format(to_file))
        report_file = to_file
    else:
        report_file = '.sos/report.md'
    #
    content = ''
    if script is not None:
        content = script
    if from_file is not None:
        try:
            with open(from_file) as rep:
                content += rep.read().decode()
        except Exception as e:
            raise RuntimeError('Failed to import report from {}: {}'.format(from_file, e))
    #
    # write report file (the ${} expressions must have been interpolated.
    if report_file == '__STDERR__':
        sys.stderr.write(content)
    else:
        with open(report_file, mode)as md:
            md.write(content)


@SoS_Action(run_mode=['run', 'interactive'])
def pandoc(script=None, output=None, **kwargs):
    '''This action can be used in three ways

    pandoc:   outputfile='report.html'
      script

    pandoc(filename='report.sos', outputfile='report.html')

    pandoc(outputfile='report.html')

    '''
    check_command('pandoc')
    #
    # in run mode, collect report and call pandoc
    sos_script = env.sos_dict['__step_context__'].filename
    # this is the case for stirng input (test only)
    if sos_script is None:
        sos_script = 'string_input'
    if script is not None:
        # get a temporary file with the content
        script_file = '{}.md'.format(os.path.basename(sos_script))
        with open(script_file, 'w') as report:
            report.write(script)
    elif 'filename' in kwargs:
        script_file = kwargs['filename']
    elif env.run_mode == 'interactive' and '__summary_report__' in env.sos_dict:
        script_file = env.sos_dict['__summary_report__']
    else:
        step_reports = glob.glob(os.path.join('.sos', 'report', '*'))
        step_reports.sort(key=natural_keys)
        # merge the files
        script_file = '{}.md'.format(os.path.basename(sos_script))
        env.logger.trace('Gathering reports {} to {}'.format(', '.join(step_reports), script_file))
        with open(script_file, 'w') as combined:
            for step_report in step_reports:
                with open(step_report, 'r') as md:
                    combined.write(md.read())
# $ pandoc -h
#
#     # this is output format
#     pandoc [OPTIONS] [FILES]
# Input formats:  commonmark, docbook, docx, epub, haddock, html, json*, latex,
#                 markdown, markdown_github, markdown_mmd, markdown_phpextra,
#                 markdown_strict, mediawiki, native, odt, opml, org, rst, t2t,
#                 textile, twiki
#                 [ *only Pandoc's JSON version of native AST]
# Output formats: asciidoc, beamer, commonmark, context, docbook, docx, dokuwiki,
#                 dzslides, epub, epub3, fb2, haddock, html, html5, icml, json*,
#                 latex, man, markdown, markdown_github, markdown_mmd,
#                 markdown_phpextra, markdown_strict, mediawiki, native, odt,
#                 opendocument, opml, org, pdf**, plain, revealjs, rst, rtf, s5,
#                 slideous, slidy, tei, texinfo, textile
#                 [**for pdf output, use latex or beamer and -o FILENAME.pdf]
# Options:
#   -f FORMAT, -r FORMAT  --from=FORMAT, --read=FORMAT
#   -t FORMAT, -w FORMAT  --to=FORMAT, --write=FORMAT
#   -o FILENAME           --output=FILENAME
#                         --data-dir=DIRECTORY
#   -R                    --parse-raw
#   -S                    --smart
#
# IGNORED
#
    arg_from = '--from={}'.format(kwargs['from']) if 'from' in kwargs else ''
    arg_to = '--to={}'.format(kwargs['to']) if 'to' in kwargs else ''
    if output is None:
        raise RuntimeError('Parameter output is required for action pandoc')
    elif not isinstance(output, str):
        raise RuntimeError('A filename is expected, {} provided'.format(output))
    arg_output = '--output={}'.format(shlex.quote(output))
    extra_args = ''
    if 'extra_args' in kwargs:
        ea = kwargs['extra_args']
        if isinstance(ea, str):
            extra_args = ea
        elif isinstance(ea, Sequence):
            extra_args = ' '.join(list(ea))
        elif isinstance(ea, dict):
            extra_args = ' '.join('--{}={}'.format(k,v) for k,v in ea.items())
    #
    command = 'pandoc {} {} {} {} {{}}'.format(arg_from, arg_to, arg_output, extra_args)
    try:
        cmd = command.replace('{}', shlex.quote(script_file))
        env.logger.trace('Running command "{}"'.format(cmd))
        if env.run_mode == 'interactive':
            # need to catch output and send to python output, which will in trun be hijacked by SoS notebook
            p = subprocess.Popen(cmd, shell=True, stderr=subprocess.PIPE, stdout=subprocess.PIPE)
            pid = p.pid
            env.register_process(p.pid, 'Runing {}'.format(script_file))
            out, err = p.communicate()
            sys.stdout.write(out.decode())
            sys.stderr.write(err.decode())
            ret = p.returncode
        else:
            p = subprocess.Popen(cmd, shell=True)
            pid = p.pid
            env.register_process(pid, 'Runing {}'.format(script_file))
            ret = p.wait()
    except Exception as e:
        env.logger.error(e)
    finally:
        env.deregister_process(p.pid)
    if ret != 0:
        temp_file = os.path.join('.sos', '{}_{}.md'.format('pandoc', os.getpid()))
        shutil.copyfile(script_file, temp_file)
        cmd = command.replace('{}', shlex.quote(temp_file))
        raise RuntimeError('Failed to execute script. The script is saved to {}. Please use command "{}" to test it.'
            .format(temp_file, cmd))
    env.logger.info('Report saved to {}'.format(output))


@SoS_Action(run_mode=['prepare', 'run', 'interactive'])
def Rmarkdown(script=None, output_file=None, **kwargs):
    '''This action can be used in three ways

    Rmarkdown:   output_file='report.html'
      script

    Rmarkdown(filename='report.sos', output_file='report.html')

    Rmarkdown(output_file='report.html')

    '''
    # if in prepare mode, check for Rmarkdown command
    if env.run_mode == 'prepare':
        return check_R_library('knitr') + check_R_library('rmarkdown')
    #
    # in run mode, collect report and call Rmarkdown
    sos_script = env.sos_dict['__step_context__'].filename
    # this is the case for stirng input (test only)
    if sos_script is None:
        sos_script = 'string_input'
    if script is not None:
        # get a temporary file with the content
        script_file = '{}.Rmd'.format(os.path.basename(sos_script))
        with open(script_file, 'w') as report:
            report.write(script)
    elif 'filename' in kwargs:
        script_file = kwargs['filename']
    elif env.run_mode == 'interactive' and '__summary_report__' in env.sos_dict:
        script_file = env.sos_dict['__summary_report__']
    else:
        step_reports = glob.glob(os.path.join('.sos', 'report', '*'))
        step_reports.sort(key=natural_keys)
        # merge the files
        script_file = '{}.Rmd'.format(os.path.basename(sos_script))
        env.logger.trace('Gathering reports {} to {}'.format(', '.join(step_reports), script_file))
        with open(script_file, 'w') as combined:
            for step_report in step_reports:
                with open(step_report, 'r') as md:
                    combined.write(md.read())
    #
    arg_output_format = ', output_format={}'.format(kwargs['output_format']) if 'output_format' in kwargs else ''
    if output_file is None:
        raise RuntimeError('Parameter output_file is required for action Rmarkdown')
    elif not isinstance(output_file, str):
        raise RuntimeError('A filename is expected, {} provided'.format(output_file))
    extra_args = ''
    if 'extra_args' in kwargs:
        ea = kwargs['extra_args']
        if isinstance(ea, str):
            extra_args = ea
        elif isinstance(ea, Sequence):
            extra_args = ' '.join(list(ea))
        elif isinstance(ea, dict):
            extra_args = ' '.join('{}={:r}'.format(k,v) for k,v in ea.items())
        extra_args = ', ' + extra_args
    #
    #   render(input, output_format = NULL, output_file = NULL, output_dir = NULL,
    #        output_options = NULL, intermediates_dir = NULL,
    #        runtime = c("auto", "static", "shiny"),
    #        clean = TRUE, params = NULL, knit_meta = NULL, envir = parent.frame(),
    #        run_pandoc = TRUE, quiet = FALSE, encoding = getOption("encoding"))
    command = '''Rscript -e "rmarkdown::render({{}}, output_file={!r} {} {})" '''.format(
        os.path.abspath(output_file), arg_output_format, extra_args)
    try:
        cmd = command.replace('{}', '{!r}'.format(script_file))
        if env.run_mode == 'interactive':
            # need to catch output and send to python output, which will in trun be hijacked by SoS notebook
            p = subprocess.Popen(cmd, shell=True, stderr=subprocess.PIPE, stdout=subprocess.PIPE)
            pid = p.pid
            env.register_process(p.pid, 'Runing {}'.format(script_file))
            out, err = p.communicate()
            sys.stdout.write(out.decode())
            sys.stderr.write(err.decode())
            ret = p.returncode
            summarizeExecution(pid)
        else:
            p = subprocess.Popen(cmd, shell=True)
            pid = p.pid
            env.register_process(pid, 'Runing {}'.format(script_file))
            ret = p.wait()
            summarizeExecution(pid)
    except Exception as e:
        env.logger.error(e)
    finally:
        env.deregister_process(p.pid)
        # os.remove(script_file)
    if ret != 0:
        temp_file = os.path.join('.sos', 'R_{}.Rmd'.format(os.getpid()))
        shutil.copyfile(script_file, temp_file)
        cmd = command.replace('{}', '{!r}'.format(temp_file))
        raise RuntimeError('Failed to execute script. The script is saved to {}. Please use command "{}" to test it.'
            .format(temp_file, cmd))
    env.logger.info('Report saved to {}'.format(output_file))




#!/usr/bin/env python3
#
# This file is part of Script of Scripts (sos), a workflow system
# for the execution of commands and scripts in different languages.
# Please visit https://github.com/bpeng2000/SOS
#
# Copyright (C) 2016 Bo Peng (bpeng@mdanderson.org)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

import os
import sys
import argparse
import webbrowser

from io import StringIO
from textwrap import dedent

from pygments import highlight
from pygments.token import Name, Keyword
from pygments.lexers import PythonLexer, TextLexer, get_lexer_by_name, guess_lexer
from pygments.formatters import HtmlFormatter, TerminalFormatter
from pygments.styles import get_all_styles
from pygments.util import shebang_matches

import nbformat
from nbformat.v4 import new_code_cell, new_markdown_cell, new_notebook
from nbconvert.exporters import Exporter

from .utils import env, pretty_size
from .actions import get_actions
from .sos_syntax import SOS_INPUT_OPTIONS, SOS_OUTPUT_OPTIONS, SOS_DEPENDS_OPTIONS, \
    SOS_RUNTIME_OPTIONS, SOS_SECTION_OPTIONS, SOS_SECTION_HEADER, SOS_CELL_LINE

__all__ = ['SoS_Lexer', 'SoS_Exporter']

#
# subcommmand show
#
# CSS style for the output of html format, extracted from github :-)
inline_css = '''
html {
  font-family: sans-serif;
  -ms-text-size-adjust: 100%;
  -webkit-text-size-adjust: 100%;
}

body {
  margin: 0;
}

a {
  background-color: transparent;
  -webkit-text-decoration-skip: objects;
}

a:active,
a:hover {
  outline-width: 0;
}

img {
  border-style: none;
}

pre {
  font-family: monospace, monospace;
  font-size: 1em;
}

button::-moz-focus-inner,
[type="button"]::-moz-focus-inner,
[type="reset"]::-moz-focus-inner,
[type="submit"]::-moz-focus-inner {
  border-style: none;
  padding: 0;
}

button:-moz-focusring,
[type="button"]:-moz-focusring,
[type="reset"]:-moz-focusring,
[type="submit"]:-moz-focusring {
  outline: 1px dotted ButtonText;
}

table {
  border-spacing: 0;
  border-collapse: collapse;
}

td {
  padding: 0;
}

* {
  box-sizing: border-box;
}

body {
  font: 13px / 1.4 Helvetica, arial, nimbussansl, liberationsans, freesans, clean, sans-serif, "Apple Color Emoji", "Segoe UI Emoji", "Segoe UI Symbol";
  color: #333;
  background-color: #fff;
}

a {
  color: #4078c0;
  text-decoration: none;
}

a:hover,
a:active {
  text-decoration: underline;
}

pre {
  margin-top: 0;
  margin-bottom: 0;
  font: 12px Consolas, "Liberation Mono", Menlo, Courier, monospace;
}

.container {
  width: 980px;
  margin-right: auto;
  margin-left: auto;
}

.container::before {
  display: table;
  content: "";
}

.container::after {
  display: table;
  clear: both;
  content: "";
}

.right {
  float: right !important;
}

::-moz-placeholder {
  color: #999;
}

:-ms-input-placeholder {
  color: #999;
}

::placeholder {
  color: #999;
}

.form-select::-ms-expand {
  opacity: 0;
}

.avatar {
  display: inline-block;
  overflow: hidden;
  line-height: 1;
  vertical-align: middle;
  border-radius: 3px;
}

.commit-tease {
  position: relative;
  padding: 10px;
  margin-bottom: -1px;
  line-height: 20px;
  color: #68777d;
  background-color: #f2f9fc;
  border: 1px solid #c9e6f2;
  border-radius: 3px;
  border-bottom-right-radius: 0;
  border-bottom-left-radius: 0;
}

.commit-tease .message {
  color: inherit;
}

.commit-tease .avatar {
  margin-top: -1px;
}

.commit-tease-sha {
  display: inline-block;
  font-family: Consolas, "Liberation Mono", Menlo, Courier, monospace;
  color: #445055;
}

.full-commit .btn-outline:not(:disabled):hover {
  color: #4078c0;
  border: 1px solid #4078c0;
}

.blob-wrapper {
  overflow-x: auto;
  overflow-y: hidden;
  border-bottom-right-radius: 3px;
  border-bottom-left-radius: 3px;
}

.blob-num {
  width: 1%;
  min-width: 50px;
  padding-right: 10px;
  padding-left: 10px;
  font-family: Consolas, "Liberation Mono", Menlo, Courier, monospace;
  font-size: 12px;
  line-height: 18px;
  color: rgba(0,0,0,0.3);
  text-align: right;
  white-space: nowrap;
  vertical-align: bottom;
  cursor: pointer;
  -webkit-user-select: none;
  -moz-user-select: none;
  -ms-user-select: none;
  user-select: none;
  border: solid #eee;
  border-width: 0 1px 0 0;
}

.blob-num:hover {
  color: rgba(0,0,0,0.6);
}

.blob-num::before {
  content: attr(data-line-number);
}

.blob-code {
  position: relative;
  padding-right: 10px;
  padding-left: 10px;
  vertical-align: top;
  white-space: nowrap;
}

.sos-comment {
  background: #f7f7f7;
}
.sos-report {
  background: #FAFAD2;
  padding-top: 5px;
}
.sos-header {
  margin-top: 20px;
  margin-bottom: 0px;
  padding-top: 5px;
  background-color: #C0C0C0;
  border-bottom-color: #dde4e6;

}
.sos-directive {
  padding-top: 5px;
  background-color: #fff9ea;
  border-color: #f1c0c0;
}
.sos-statement {
  padding-top: 5px;
  background-color: #f2f9fc;
}
.sos-error {
  background: #ffdddd;
}
.sos-script {
  position: relative;
  padding-right: 10px;
  padding-left: 30px;
  vertical-align: top;
  background-color: #eff0f1;
}


.blob-code-inner {
  overflow: visible;
  font-family: Consolas, "Liberation Mono", Menlo, Courier, monospace;
  font-size: 12px;
  color: #333;
  word-wrap: normal;
  white-space: pre;
}

.blob-code-inner::before {
  content: "";
}

body {
  min-width: 1020px;
  word-wrap: break-word;
}

.user-mention {
  font-weight: bold;
  color: #333;
  white-space: nowrap;
}

.prose-diff>.markdown-body :not(li.moved).removed {
  color: #a33;
  text-decoration: line-through;
  background: #ffeaea;
}

.prose-diff>.markdown-body :not(.github-user-ins):not(li.moved).removed {
  text-decoration: line-through;
}

.prose-diff>.markdown-body :not(li.moved).added {
  background: #eaffea;
}

.prose-diff>.markdown-body :not(.github-user-del):not(li.moved).added li:not(.moved):not(.github-user-del).added {
  text-decoration: none;
}

:checked+.radio-label {
  position: relative;
  z-index: 1;
  border-color: #4078c0;
}

.select-menu-text-filter input::-moz-placeholder {
  color: #aaa;
}

.select-menu-text-filter input:-ms-input-placeholder {
  color: #aaa;
}

.select-menu-text-filter input::placeholder {
  color: #aaa;
}

.tab-size[data-tab-size="8"] {
  -moz-tab-size: 8;
  -o-tab-size: 8;
  tab-size: 8;
}

# override source
.source {
  background: auto;
}

.file {
  position: relative;
  margin-top: 20px;
  margin-bottom: 15px;
  border: 1px solid #ddd;
  border-radius: 3px;
}

.file-header {
  padding: 5px 10px;
  background-color: #f7f7f7;
  border-bottom: 1px solid #d8d8d8;
  border-top-left-radius: 2px;
  border-top-right-radius: 2px;
}

.file-header::before {
  display: table;
  content: "";
}

.file-header::after {
  display: table;
  clear: both;
  content: "";
}

.file-actions {
  float: right;
  padding-top: 3px;
}

.file-info {
  float: left;
  font-family: Consolas, "Liberation Mono", Menlo, Courier, monospace;
  font-size: 12px;
  line-height: 32px;
}

.file-info-divider {
  display: inline-block;
  width: 1px;
  height: 18px;
  margin-right: 3px;
  margin-left: 3px;
  vertical-align: middle;
  background-color: #ddd;
}

.header-search-input::-ms-clear {
  display: none;
}

.org-validate-group input::-ms-clear {
  display: none;
}

'''

template_pre_style = '''
<!DOCTYPE html PUBLIC "-//W3C//DTD HTML 4.01//EN" "http://www.w3.org/TR/html4/strict.dtd">
<html>
<head>
    <meta charset='utf-8'>



 <title>%s</title>
<meta http-equiv="content-type" content="text/html; charset=None">
<style type="text/css">
'''

template_pre_table = '''

 </style>
</head>
<body>


<div class="container new-discussion-timeline experiment-repo-nav">

 <div class="commit-tease">
      <span class="right">
      </span>
 <div>
        <a class="header-logo-invertocat" href="https://github.com/BoPeng/SOS" data-hotkey="g d" aria-label="Homepage">
        <svg xmlns="http://www.w3.org/2000/svg" style="width: 25px; height: 25px" viewBox="0 0 64 64" enable-background="new 0 0 64 64"><path fill="#ec1c24" d="M63.791,56.913c0,3.802-3.076,6.882-6.876,6.882H6.874C3.078,63.795,0,60.715,0,56.913V6.877    C0,3.08,3.078,0,6.874,0h50.042c3.8,0,6.876,3.08,6.876,6.877V56.913z"/><path fill="#c32129" d="m0 48.574v8.339c0 3.802 3.078 6.882 6.874 6.882h50.04c3.8 0 6.876-3.08 6.876-6.882v-50.04c-10.894 39.569-50.05 42.08-63.792 41.697"/><g fill="#fff"><path d="m20.445 36.649c0 1.311-.337 2.49-1.015 3.536-.678 1.051-1.667 1.869-2.973 2.463-1.305.59-2.849.888-4.636.888-2.146 0-3.911-.405-5.306-1.215-.986-.584-1.789-1.357-2.404-2.335-.617-.973-.927-1.919-.927-2.836 0-.532.186-.989.553-1.371.374-.38.844-.568 1.418-.568.462 0 .856.148 1.178.444.319.296.592.733.821 1.319.278.689.574 1.268.896 1.729.321.467.776.849 1.356 1.148.582.304 1.35.455 2.298.455 1.305 0 2.361-.304 3.174-.911.815-.607 1.227-1.368 1.227-2.277 0-.719-.221-1.303-.66-1.754-.439-.448-1.01-.791-1.705-1.028-.697-.239-1.628-.492-2.792-.758-1.56-.365-2.867-.792-3.919-1.28-1.054-.49-1.885-1.156-2.502-2-.619-.844-.929-1.893-.929-3.148 0-1.195.325-2.255.979-3.186.65-.927 1.593-1.64 2.828-2.14 1.238-.498 2.688-.749 4.357-.749 1.334 0 2.485.167 3.457.496.974.333 1.783.772 2.426 1.319.643.548 1.111 1.126 1.405 1.726.3.602.447 1.191.447 1.764 0 .523-.186.993-.554 1.417-.375.417-.835.629-1.387.629-.502 0-.888-.127-1.148-.38-.265-.249-.545-.663-.853-1.234-.396-.823-.866-1.462-1.422-1.918-.555-.46-1.44-.692-2.667-.692-1.136 0-2.051.251-2.747.749-.693.5-1.046 1.102-1.046 1.799 0 .439.121.811.356 1.127.237.317.562.591.978.817.416.227.835.406 1.26.535.425.125 1.128.312 2.104.561 1.223.286 2.334.602 3.323.948.996.345 1.838.766 2.536 1.26.696.494 1.238 1.117 1.63 1.875.389.755.585 1.68.585 2.777"/><path d="m31.57 21.08c2.249 0 4.184.457 5.8 1.372 1.614.915 2.835 2.212 3.664 3.895.829 1.686 1.246 3.663 1.246 5.933 0 1.682-.228 3.208-.68 4.579-.457 1.372-1.137 2.563-2.048 3.572-.907 1.01-2.023 1.775-3.347 2.309-1.324.534-2.841.801-4.548.801-1.699 0-3.223-.274-4.565-.82-1.344-.549-2.465-1.323-3.36-2.322-.899-.995-1.579-2.194-2.038-3.6-.457-1.4-.69-2.915-.69-4.546 0-1.667.241-3.202.719-4.595.48-1.393 1.172-2.577 2.081-3.554.905-.98 2.02-1.727 3.319-2.246 1.304-.52 2.786-.778 4.447-.778m6.253 11.172c0-1.593-.257-2.968-.772-4.136-.516-1.164-1.246-2.045-2.202-2.643-.952-.598-2.044-.896-3.278-.896-.88 0-1.693.167-2.439.496-.743.331-1.385.811-1.924 1.444-.539.631-.964 1.44-1.275 2.424-.312.981-.466 2.086-.466 3.311 0 1.232.155 2.352.466 3.345.312.997.75 1.826 1.318 2.484.568.657 1.223 1.146 1.958 1.472.737.329 1.542.49 2.42.49 1.127 0 2.161-.28 3.105-.845.944-.562 1.695-1.431 2.252-2.608.555-1.173.837-2.619.837-4.338"/><path d="m60.38 36.649c0 1.311-.339 2.49-1.016 3.536-.676 1.051-1.665 1.869-2.97 2.463-1.304.59-2.85.888-4.637.888-2.146 0-3.91-.405-5.302-1.215-.989-.584-1.793-1.357-2.41-2.335-.615-.973-.925-1.919-.925-2.836 0-.532.185-.989.553-1.371.372-.38.845-.568 1.416-.568.465 0 .858.148 1.18.444.319.296.594.733.821 1.319.274.689.576 1.268.894 1.729.323.467.772.849 1.357 1.148.582.304 1.347.455 2.299.455 1.305 0 2.361-.304 3.174-.911.815-.607 1.225-1.368 1.225-2.277 0-.719-.219-1.303-.66-1.754-.438-.448-1.01-.791-1.7-1.028-.7-.239-1.629-.492-2.794-.758-1.561-.365-2.867-.792-3.921-1.28-1.053-.49-1.884-1.156-2.503-2-.619-.844-.927-1.893-.927-3.148 0-1.195.325-2.255.98-3.186.648-.927 1.591-1.64 2.829-2.14 1.232-.498 2.687-.749 4.354-.749 1.337 0 2.486.167 3.463.496.972.333 1.778.772 2.422 1.319.639.548 1.108 1.126 1.402 1.726.302.602.447 1.191.447 1.764 0 .523-.185.993-.553 1.417-.374.417-.835.629-1.388.629-.501 0-.888-.127-1.148-.38-.264-.249-.546-.663-.85-1.234-.398-.823-.866-1.462-1.427-1.918-.553-.46-1.44-.692-2.665-.692-1.137 0-2.054.251-2.751.749-.691.5-1.04 1.102-1.04 1.799 0 .439.115.811.356 1.127.235.317.561.591.974.817.419.227.837.406 1.262.535.424.125 1.129.312 2.101.561 1.225.286 2.336.602 3.327.948.991.345 1.838.766 2.535 1.26.696.494 1.239 1.117 1.627 1.875.394.755.589 1.68.589 2.777"/></g> </svg>
        </a>

        <span class="user-mention">%s</span>
      </div>
      </div>

  <div class="repository-content">
  <div class="file">
  <div class="file-header">

  <div class="file-info">
      %s
      <span class="file-info-divider"></span>
    %s
  </div>
  <div class="file-actions">
  </div>
  </div>
 <div itemprop="text" class="blob-wrapper data type-python">

'''

template_post_table = '''
</div>
</div>
</div>
</body></html>
'''


class SoS_Lexer(PythonLexer):
    """
    A Python lexer with SOS keywords, used for SoS directive
    """

    name = 'Sript of Scripts'
    aliases = ['sos']

    # override the mimetypes to not inherit them from python
    mimetypes = []
    filenames = ['*.sos']
    mimetypes = ['text/x-sos', 'application/x-sos']

    PythonLexer.tokens['root'].insert(0, (r'(^\w+)\s*:', Keyword.Namespace))

    EXTRA_KEYWORDS = set( SOS_INPUT_OPTIONS + SOS_OUTPUT_OPTIONS +
        SOS_DEPENDS_OPTIONS + SOS_RUNTIME_OPTIONS + SOS_SECTION_OPTIONS +
        get_actions())

    def get_tokens_unprocessed(self, text):
        for index, token, value in \
                PythonLexer.get_tokens_unprocessed(self, text):
            if token is Name and value in self.EXTRA_KEYWORDS:
                yield index, Keyword.Pseudo, value
            else:
                yield index, token, value

    def analyse_text(text):
        return (shebang_matches(text, r'sos-runner') or \
            '#fileformat=SOS' in text[:1000])

class ContinuousHtmlFormatter(HtmlFormatter):
    def _wrap_tablelinenos(self, inner):
        # change wrap_table to do not output '<table' and '</table>' because
        # we will output several tables and use a single table tag around them.
        dummyoutfile = StringIO()
        lncount = 0
        for t, line in inner:
            if t:
                lncount += 1
            dummyoutfile.write(line)

        fl = self.linenostart
        mw = len(str(lncount + fl - 1))
        sp = self.linenospecial
        st = self.linenostep
        la = self.lineanchors
        aln = self.anchorlinenos
        nocls = self.noclasses
        if sp:
            lines = []

            for i in range(fl, fl+lncount):
                if i % st == 0:
                    if i % sp == 0:
                        if aln:
                            lines.append('<a href="#%s-%d" class="special">%*d</a>' %
                                         (la, i, mw, i))
                        else:
                            lines.append('<span class="special">%*d</span>' % (mw, i))
                    else:
                        if aln:
                            lines.append('<a href="#%s-%d">%*d</a>' % (la, i, mw, i))
                        else:
                            lines.append('%*d' % (mw, i))
                else:
                    lines.append('')
            ls = '\n'.join(lines)
        else:
            lines = []
            for i in range(fl, fl+lncount):
                if i % st == 0:
                    if aln:
                        lines.append('<a href="#%s-%d">%*d</a>' % (la, i, mw, i))
                    else:
                        lines.append('%*d' % (mw, i))
                else:
                    lines.append('')
            ls = '\n'.join(lines)

        # in case you wonder about the seemingly redundant <div> here: since the
        # content in the other cell also is wrapped in a div, some browsers in
        # some configurations seem to mess up the formatting...
        if nocls:
            yield 0, ('<tr><td><div class="linenodiv" '
                      'style="background-color: #f0f0f0; padding-right: 10px">'
                      '<pre style="line-height: 125%">' +
                      ls + '</pre></div></td><td class="code">')
        else:
            yield 0, ('<tr><td class="blob-num js-line-number"><div class="linenodiv"><pre>' +
                      ls + '</pre></div></td><td class="code">')
        yield 0, dummyoutfile.getvalue()
        yield 0, '</td></tr>'
        self.linenostart += lncount

def write_html_content(content_type, content, formatter, html):
    # dedent content but still keeps empty lines
    old_class = formatter.cssclass
    nlines = len(content)
    content = dedent(''.join(content))
    # ' ' to keep pygments from removing empty lines
    # split, merge by \n can introduce one additional line
    content = [' \n' if x == '' else x + '\n' for x in content.split('\n')][:nlines]
    #
    if content_type == 'COMMENT':
        formatter.cssclass = 'source blob-code sos-comment'
        html.write('{}\n'.format(highlight(''.join(content),
            SoS_Lexer(), formatter)))
    elif content_type in ('REPORT', 'report'):
        formatter.cssclass = 'source blob-code sos-report'
        html.write('{}\n'.format(highlight(''.join(content),
            TextLexer(), formatter)))
    elif content_type == 'SECTION':
        formatter.cssclass = 'source blob-code sos-header'
        html.write('{}\n'.format(highlight(''.join(content),
            SoS_Lexer(), formatter)))
    elif content_type == 'DIRECTIVE':
        formatter.cssclass = 'source blob-code sos-directive'
        html.write('{}\n'.format(highlight(''.join(content),
            SoS_Lexer(), formatter)))
    elif content_type == 'ASSIGNMENT':
        formatter.cssclass = 'source blob-code sos-statement'
        html.write('{}\n'.format(highlight(''.join(content),
            SoS_Lexer(), formatter)))
    elif content_type == 'STATEMENT':
        formatter.cssclass = 'source blob-code sos-statement'
        html.write('{}\n'.format(highlight(''.join(content),
            SoS_Lexer(), formatter)))
    elif content_type == 'ERROR':
        formatter.cssclass = 'source blob-code sos-error '
        html.write('{}\n'.format(highlight(''.join(content),
            SoS_Lexer(), formatter)))
    else:
        formatter.cssclass = 'source blob-code sos-script '
        if content_type == 'run':
            content_type = 'bash'
        elif content_type == 'node':
            content_type = 'JavaScript'
        elif content_type == 'report':
            content_type == 'text'
        try:
            lexer = get_lexer_by_name(content_type)
        except:
            try:
                lexer = guess_lexer(''.join(content))
            except:
                lexer = TextLexer()
        html.write('{}\n'.format(highlight((''.join(content)),
            lexer, formatter)))
    formatter.cssclass = old_class

#
# Converter to HTML
#
#
def parse_html_args(style_args):
    parser = argparse.ArgumentParser()
    parser.add_argument('--raw')
    parser.add_argument('--style', choices=list(get_all_styles()), default='default')
    parser.add_argument('--linenos', action='store_true')
    try:
        args = parser.parse_args(style_args)
    except Exception as e:
        raise RuntimeError('Unrecognized style argument {}: {}'
            .format(' '.join(style_args), e))
    return vars(args)

def script_to_html(transcript, script_file, html_file, style_args):
    '''
    Write a HTML file with the transcript of a SOS file.
    '''
    if html_file == '__BROWSER__':
        html_file = os.path.join('.sos', '{}.html'.format(os.path.basename(script_file)))
    sargs = parse_html_args(style_args)
    formatter = ContinuousHtmlFormatter(cssclass="source", full=False,
        **{x:y for x,y in sargs.items() if x != 'raw'})
    with open(html_file, 'w') as html:
        html.write(template_pre_style % os.path.basename(script_file))
        html.write(inline_css)
        # remove background definition so that we can use our own
        html.write('\n'.join(x for x in formatter.get_style_defs().split('\n') if 'background' not in x))
        if 'raw' in sargs and sargs['raw']:
            raw_link = '<a href="{}" class="commit-tease-sha">{}</a>'.format(sargs['raw'], script_file)
        else:
            raw_link = script_file
        html.write(template_pre_table % (os.path.basename(script_file), raw_link,
            pretty_size(os.path.getsize(script_file))))
        #
        html.write('<table class="highlight tab-size js-file-line-container">')
        with open(transcript) as script:
            content = []
            content_type = None
            content_number = None
            next_type = None
            for line in script:
                line_type, line_no, script_line = line.split('\t', 2)
                # Does not follow section because it has to be one line
                if line_type == 'FOLLOW' and content_type in (None, 'SECTION'):
                    line_type = 'COMMENT'
                if content_type == line_type or line_type == 'FOLLOW':
                    if next_type is not None and not script_line.rstrip().endswith(','):
                        formatter.linenostart = content_number
                        write_html_content(content_type, content, formatter, html)
                        content = [script_line]
                        content_type = next_type
                        content_number = int(line_no)
                        next_type = None
                    else:
                        content.append(script_line)
                else:
                    if content:
                        formatter.linenostart = content_number
                        write_html_content(content_type, content, formatter, html)
                    if line_type.startswith('SCRIPT_'):
                        content_type = 'DIRECTIVE'
                        next_type = line_type[7:]
                    else:
                        content_type = line_type
                    content_number = int(line_no)
                    content = [script_line]
        if content:
            formatter.linenostart = content_number
            write_html_content(content_type, content, formatter, html)
        html.write('</table>')
        html.write(template_post_table)
    #
    if html_file.startswith('.sos'):
        url = 'file://{}'.format(os.path.abspath(html_file))
        env.logger.info('Viewing {} in a browser'.format(url))
        webbrowser.open(url, new=2)
    else:
        env.logger.info('SoS script saved to {}'.format(html_file))


def workflow_to_html(workflow, script_file, html_file, style_args):
    '''
    Write a HTML file with the transcript of a SOS file.
    '''
    if html_file == '__BROWSER__':
        html_file = os.path.join('.sos', '{}.html'.format(os.path.basename(script_file)))
    sargs = parse_html_args(style_args)
    formatter = ContinuousHtmlFormatter(cssclass="source", full=False,
        **{x:y for x,y in sargs.items() if x != 'raw'})
    with open(html_file, 'w') as html:
        html.write(template_pre_style % ('{} - {}'.format(os.path.basename(script_file), workflow)))
        html.write(inline_css)
        # remove background definition so that we can use our own
        html.write('\n'.join(x for x in formatter.get_style_defs().split('\n') if 'background' not in x))
        if 'raw' in sargs and sargs['raw']:
            raw_link = '<a href="{}" class="commit-tease-sha">{}</a>'.format(sargs['raw'], script_file)
        else:
            raw_link = script_file
        html.write(template_pre_table % (os.path.basename(script_file), raw_link, ''))
        #
        html.write('<table class="highlight tab-size js-file-line-container">')
        if workflow.sections and workflow.sections[0].global_def:
            write_html_content('STATEMENT', workflow.sections[0].global_def, formatter, html)
        #
        for section in workflow.sections:
            write_html_content('SECTION', '[{}_{}]'.format(section.name, section.index), formatter, html)
            #
            if section.comment:
                write_html_content('COMMENT', '#' + section.comment, formatter, html)
            for stmt in section.statements:
                if stmt[0] == ':':
                    write_html_content('DIRECTIVE', '{} : {}'.format(stmt[1], stmt[2]), formatter, html)
                elif stmt[0] == '=':
                    write_html_content('STATEMENT', '{} = {}'.format(stmt[1], stmt[2]), formatter, html)
                else:
                    write_html_content('STATEMENT', stmt[1].strip(), formatter, html)
            if section.task:
                #write_html_content('DIRECTIVE', 'task:', formatter, html)
                write_html_content('STATEMENT', section.task.strip(), formatter, html)
        html.write('</table>')
        html.write(template_post_table)
    #
    if html_file.startswith('.sos'):
        url = 'file://{}'.format(os.path.abspath(html_file))
        env.logger.info('Viewing {} in a browser'.format(url))
        webbrowser.open(url, new=2)
    else:
        env.logger.info('Workflow saved to {}'.format(html_file))

#
# Converter to Markdown
#
#

def markdown_content(content_type, content, fh):
    # write content to a file
    import re
    if content_type == 'COMMENT':
        fh.write('{}\n'.format(''.join([x.lstrip('#').strip() + '  \n'
                                        for x in content if not x.startswith('#!')])))
    elif content_type == 'REPORT':
        fh.write('{}\n'.format(''.join(content)))
    elif content_type == 'SECTION':
        fh.write('## {}\n'.format(''.join(content)))
    elif content_type == 'DIRECTIVE':
        fh.write('{}\n'.format(''.join(['**{}**  \n'.format(re.sub(r'(\$|_)', r'`\1`', x).strip())
                                        for x in content])))
    elif content_type == 'ASSIGNMENT':
        fh.write('```python\n{}\n```\n'.format(''.join(content))),
    elif content_type == 'STATEMENT':
        fh.write('```python\n{}\n```\n'.format(''.join(content))),
    elif content_type == 'ERROR':
        fh.write('{}\n'.format(''.join(content))),
    else:
        if content_type == 'run':
            content_type = 'bash'
        elif content_type == 'node':
            content_type = 'JavaScript'
        elif content_type == 'report':
            content_type == ''
        fh.write('```{}\n{}```\n'.format(content_type, ''.join(content)))

def script_to_markdown(transcript, script_file, markdown_file):
    '''
    Write a markdown file with the transcript of a SOS file.
    '''
    if markdown_file == '__STDOUT__':
        markdown = sys.stdout
    else:
        markdown = open(markdown_file, 'w')
    # remove background definition so that we can use our own
    with open(transcript) as script:
        content = []
        content_type = None
        # content_number = None
        next_type = None
        for line in script:
            line_type, line_no, script_line = line.split('\t', 2)
            # Does not follow section because it has to be one line
            if line_type == 'FOLLOW' and content_type in (None, 'SECTION'):
                line_type = 'COMMENT'
            if content_type == line_type or line_type == 'FOLLOW':
                if next_type is not None and not script_line.rstrip().endswith(','):
                    markdown_content(content_type, content, markdown)
                    content = [script_line]
                    content_type = next_type
                    # content_number = int(line_no)
                    next_type = None
                else:
                    content.append(script_line)
            else:
                if content:
                    markdown_content(content_type, content, markdown)
                if line_type.startswith('SCRIPT_'):
                    content_type = 'DIRECTIVE'
                    next_type = line_type[7:]
                else:
                    content_type = line_type
                # content_number = int(line_no)
                content = [script_line]
    if content:
        markdown_content(content_type, content, markdown)
    if markdown != sys.stdout:
        env.logger.info('SoS script saved to {}'.format(markdown_file))


def workflow_to_markdown(workflow, script_file, markdown_file):
    '''
    Write a markdown file with the transcript of a SOS file.
    '''
    if markdown_file == '__STDOUT__':
        markdown = sys.stdout
    else:
        markdown = open(markdown_file, 'w')
    if workflow.sections and workflow.sections[0].global_def:
        markdown_content('STATEMENT', workflow.sections[0].global_def, markdown)
    #
    for section in workflow.sections:
        markdown_content('SECTION', '[{}_{}]'.format(section.name, section.index), markdown)
        #
        if section.comment:
            markdown_content('COMMENT', '#' + section.comment, markdown)
        for stmt in section.statements:
            if stmt[0] == ':':
                markdown_content('DIRECTIVE', '{} : {}'.format(stmt[1], stmt[2]), markdown)
            elif stmt[0] == '=':
                markdown_content('STATEMENT', '{} = {}'.format(stmt[1], stmt[2]), markdown)
            else:
                markdown_content('STATEMENT', stmt[1].strip(), markdown)
        if section.task:
            markdown_content('STATEMENT', section.task.strip(), markdown)
        markdown_content('COMMENT', '\n', markdown)
    #
    if markdown != sys.stdout:
        env.logger.info('Workflow saved to {}'.format(markdown_file))


#
# Converter from Notebook
#

def parse_convert_args(convert_args):
    parser = argparse.ArgumentParser()
    parser.add_argument('--reorder', action='store_true')
    parser.add_argument('--reset-index', action='store_true')
    parser.add_argument('--add-header', action='store_true')
    parser.add_argument('--no-index', action='store_true')
    parser.add_argument('--remove-magic', action='store_true')
    parser.add_argument('--md-to-report', action='store_true')
    try:
        args = parser.parse_args(convert_args)
    except Exception as e:
        raise RuntimeError('Unrecognized style argument {}: {}'
            .format(' '.join(convert_args), e))
    return args


# This class cannot be defined in .kernel because it would cause some
# weird problem with unittesting not able to resolve __main__
class SoS_Exporter(Exporter):
    def __init__(self, config=None, reorder=False, reset_index=False, add_header=False,
            no_index=False, remove_magic=False, md_to_report=False,
            **kwargs):
        self.reorder = reorder
        self.reset_index = reset_index
        self.add_header = add_header
        self.no_index = no_index
        self.remove_magic = remove_magic
        self.md_to_report = md_to_report
        self.output_extension = '.sos'
        self.output_mimetype = 'text/x-sos'
        Exporter.__init__(self, config, **kwargs)

    def from_notebook_cell(self, cell, fh, idx = 0):
        if not hasattr(cell, 'execution_count') or cell.execution_count is None or self.no_index:
            fh.write('\n%cell {}\n'.format(cell.cell_type))
        else:
            idx += 1
            fh.write('\n%cell {} {}\n'.format(cell.cell_type,
                                              idx if self.reset_index else cell.execution_count))
        if cell.cell_type == 'code':
            if any(cell.source.startswith(x) for x in ('%run', '%restart', '%dict', '%use', '%with', '%set', '%paste')):
                if self.remove_magic:
                    cell.source = '\n'.join(cell.source.split('\n')[1:])
            if self.add_header and not any([SOS_SECTION_HEADER.match(x) for x in cell.source.split('\n')]):
                cell.source = '[{}]\n'.format(idx if self.reset_index else cell.execution_count) + cell.source
            fh.write(cell.source.strip() + '\n')
        elif cell.cell_type == "markdown":
            fh.write('\n'.join('#! ' + x for x in cell.source.split('\n') if x.strip()) + '\n')
        return idx

    def from_notebook_node(self, nb, resources, **kwargs):
        #
        if self.reorder:
            unnumbered_cells = {x: y for x, y in enumerate(nb.cells)
                              if not hasattr(y, 'execution_count') or y.execution_count is None}
            numbered_cells = [y for y in nb.cells
                              if hasattr(y, 'execution_count') and y.execution_count is not None]
            numbered_cells = sorted(numbered_cells, key = lambda x: x.execution_count)
            cells = []
            for idx in range(len(nb.cells)):
                if idx in unnumbered_cells:
                    cells.append(unnumbered_cells[idx])
                else:
                    cells.append(numbered_cells.pop(0))
        else:
            cells = nb.cells
        with StringIO() as fh:
            fh.write('#!/usr/bin/env sos-runner\n')
            fh.write('#fileformat=SOSNB1.0\n')
            idx = 0
            for cell in cells:
                idx = self.from_notebook_cell(cell, fh, idx)
            content = fh.getvalue()
        resources['output_extension'] = '.sos'
        return content, resources


def notebook_to_script(notebook_file, sos_file, convert_args=[]):
    '''
    convert a ipython notebook.
    '''
    sargs = parse_convert_args(convert_args)
    exporter = SoS_Exporter(reorder=sargs.reorder, reset_index=sargs.reset_index,
                            add_header=sargs.add_header, no_index=sargs.no_index,
                            remove_magic=sargs.remove_magic, md_to_report=sargs.md_to_report)
    notebook = nbformat.read(notebook_file, nbformat.NO_CONVERT)
    output, resource = exporter.from_notebook_node(notebook, {})
    if sos_file == '__STDOUT__':
        sys.stdout.write(output)
    else:
        with open(sos_file, 'w') as sos:
            sos.write(output)
        env.logger.info('SoS script saved to {}'.format(sos_file))

#
# Converter to Notebook
#
def add_cell(cells, content, cell_type, cell_count):
    # if a section consist of all report, report it as a markdown cell
    if not content:
        return
    if cell_type not in ('code', 'markdown'):
        env.logger.warning('Unrecognized cell type {}, code assumed.'.format(cell_type))
    if cell_type == 'markdown' and any(x.strip() and not x.startswith('#! ') for x in content):
        env.logger.warning('Markdown lines not starting with #!, code cell assumed.')
        cell_type = 'code'
    #
    if cell_type == 'markdown':
        cells.append(new_markdown_cell(source=''.join([x[3:] for x in content]),
            execution_count=cell_count))
    else:
        cells.append(
             new_code_cell(
                 source=''.join(content),
                 execution_count=cell_count)
        )

def script_to_notebook(script_file, notebook_file):
    '''
    Write a notebook file with the transcript of a SOS file.
    '''
    cells = []
    cell_count = 1
    cell_type = 'code'
    content = []

    with open(script_file) as script:
        first_block = True
        for line in script:
            if line.startswith('#') and first_block:
                if line.startswith('#!'):
                    continue
                if line.startswith('#fileformat='):
                    if not line[12:].startswith('SOS'):
                        raise RuntimeError('{} is not a SoS script according to #fileformat line.'.format(script_file))
                    continue

            first_block = False

            mo = SOS_CELL_LINE.match(line)
            if mo:
                # eat a new line before the CELL_LINE
                if content and content[-1] == '\n':
                    content = content[:-1]
                # get ride of empty content
                if not any(x.strip() for x in content):
                    continue
                add_cell(cells, content, cell_type, cell_count)
                cell_count += 1

                cell_type = mo.group(1)
                content = []
                continue
            else:
                content.append(line)
    #
    if content and any(x.strip() for x in content):
        add_cell(cells, content, cell_type, cell_count)
    #
    nb = new_notebook(cells = cells,
        metadata = {
            'kernelspec' : {
                "display_name": "SoS",
                "language": "sos",
                "name": "sos"
            },
            "language_info": {
                "file_extension": ".sos",
                "mimetype": "text/x-sos",
                "name": "sos",
                "pygments_lexer": "python"
            }
        }
    )
    if notebook_file == '__STDOUT__':
        nbformat.write(nb, sys.stdout, 4)
    else:
        with open(notebook_file, 'w') as notebook:
            nbformat.write(nb, notebook, 4)
        env.logger.info('SoS script saved to {}'.format(notebook_file))


def workflow_to_notebook(workflow, script_file, notebook_file):
    '''
    Write a notebook file with the transcript of a SOS file.
    '''
    cells = []
    cell_count = 1
    if workflow.sections and workflow.sections[0].global_def:
        cells.append(
            new_code_cell(
            source = workflow.sections[0].global_def,
            execution_count=cell_count)
        )
        cell_count += 1
    #
    for section in workflow.sections:
        # FIXME: section option is ignored
        cell_src_code = '[{}_{}]\n'.format(section.name, section.index)
        #
        if section.comment:
            cell_src_code += '# ' + section.comment + '\n'
        for stmt in section.statements:
            if stmt[0] == ':':
                cell_src_code += '{}: {}\n'.format(stmt[1], stmt[2])
            elif stmt[0] == '=':
                cell_src_code += '{} = {}\n'.format(stmt[1], stmt[2])
            else:
                cell_src_code += stmt[1].strip() + '\n'
        if section.task:
            cell_src_code += 'task:\n' + section.task.strip() + '\n'
        cells.append(
            new_code_cell(
            source = cell_src_code,
            execution_count=cell_count)
        )
        cell_count += 1
    #
    nb = new_notebook(cells = cells,
        metadata = {
            'kernelspec' : {
                "display_name": "SoS",
                "language": "sos",
                "name": "sos"
            },
            "language_info": {
                "file_extension": ".sos",
                "mimetype": "text/x-sos",
                "name": "sos",
                "pygments_lexer": "python"
            }
        }
    )
    if notebook_file == '__STDOUT__':
            nbformat.write(nb, sys.stdout, 4)
    else:
        with open(notebook_file, 'w') as notebook:
            nbformat.write(nb, notebook, 4)
        #
        env.logger.info('Workflow saved to {}'.format(notebook_file))

#
# Output to terminal
#
def write_content(content_type, content, formatter, fh=sys.stdout):
    #
    nlines = len(content)
    content = dedent(''.join(content))
    # ' ' to keep pygments from removing empty lines
    # split, merge by \n can introduce one additional line
    content = [' \n' if x == '' else x + '\n' for x in content.split('\n')][:nlines]
    #
    if content_type == 'COMMENT':
        fh.write(highlight(''.join(content), SoS_Lexer(), formatter))
    elif content_type in ('REPORT', 'report'):
        fh.write(highlight(''.join(content), TextLexer(), formatter))
    elif content_type == 'SECTION':
        fh.write(highlight(''.join(content), SoS_Lexer(), formatter))
    elif content_type == 'DIRECTIVE':
        fh.write(highlight(''.join(content), SoS_Lexer(), formatter))
    elif content_type == 'ASSIGNMENT':
        fh.write(highlight(''.join(content), SoS_Lexer(), formatter))
    elif content_type == 'STATEMENT':
        fh.write(highlight(''.join(content), SoS_Lexer(), formatter))
    elif content_type == 'ERROR':
        fh.write(highlight(''.join(content), SoS_Lexer(), formatter))
    else:
        if content_type == 'run':
            content_type = 'bash'
        elif content_type == 'node':
            content_type = 'JavaScript'
        elif content_type == 'report':
            content_type == 'text'
        try:
            lexer = get_lexer_by_name(content_type)
        except:
            try:
                lexer = guess_lexer(''.join(content))
            except:
                lexer = TextLexer()
        fh.write(highlight((''.join(content)), lexer, formatter))

def parse_term_args(style_args):
    parser = argparse.ArgumentParser()
    parser.add_argument('--bg', choices=['light', 'dark'], default='light')
    parser.add_argument('--linenos', action='store_true')
    try:
        args = parser.parse_args(style_args)
    except Exception as e:
        raise RuntimeError('Unrecognized style argument {}: {}'
            .format(' '.join(style_args), e))
    return vars(args)

def script_to_term(transcript, script_file, style_args):
    '''
    Write script to terminal
    '''
    sargs = parse_term_args(style_args)
    env.logger.trace('Using style argument {}'.format(sargs))
    formatter = TerminalFormatter(**sargs)
    # remove background definition so that we can use our own
    with open(transcript) as script:
        content = []
        content_type = None
        # content_number = None
        next_type = None
        for line in script:
            line_type, line_no, script_line = line.split('\t', 2)
            # Does not follow section because it has to be one line
            if line_type == 'FOLLOW' and content_type in (None, 'SECTION'):
                line_type = 'COMMENT'
            if content_type == line_type or line_type == 'FOLLOW':
                if next_type is not None and not script_line.rstrip().endswith(','):
                    write_content(content_type, content, formatter)
                    content = [script_line]
                    content_type = next_type
                    # content_number = int(line_no)
                    next_type = None
                else:
                    content.append(script_line)
            else:
                if content:
                    write_content(content_type, content, formatter)
                if line_type.startswith('SCRIPT_'):
                    content_type = 'DIRECTIVE'
                    next_type = line_type[7:]
                else:
                    content_type = line_type
                # content_number = int(line_no)
                content = [script_line]
    if content:
        write_content(content_type, content, formatter)

def workflow_to_term(workflow, script_file, style_args):
    '''
    Write a workflow to terminal
    '''
    sargs = parse_term_args(style_args)
    env.logger.trace('Using style argument {}'.format(sargs))
    formatter = TerminalFormatter(**sargs)
    if workflow.sections and workflow.sections[0].global_def:
        write_content('STATEMENT', workflow.sections[0].global_def, formatter)
    #
    for section in workflow.sections:
        write_content('SECTION', '[{}_{}]'.format(section.name, section.index), formatter)
        #
        if section.comment:
            write_content('COMMENT', '#' + section.comment, formatter)
        for stmt in section.statements:
            if stmt[0] == ':':
                write_content('DIRECTIVE', '{} : {}'.format(stmt[1], stmt[2]), formatter)
            elif stmt[0] == '=':
                write_content('STATEMENT', '{} = {}'.format(stmt[1], stmt[2]), formatter)
            else:
                write_content('STATEMENT', stmt[1].strip(), formatter)
        if section.task:
            write_content('STATEMENT', section.task.strip(), formatter)
        write_content('COMMENT', '\n', formatter)



#!/usr/bin/env python3
#
# This file is part of Script of Scripts (sos), a workflow system
# for the execution of commands and scripts in different languages.
# Please visit https://github.com/bpeng2000/SOS
#
# Copyright (C) 2016 Bo Peng (bpeng@mdanderson.org)
##
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
import os
import sys
import yaml
import shlex
import argparse
import time
import builtins
import keyword
from collections.abc import Sequence
import multiprocessing as mp
from queue import Empty

from io import StringIO
from ._version import __version__
from .sos_step import Dryrun_Step_Executor, Prepare_Step_Executor, SP_Step_Executor, MP_Step_Executor, RQ_Step_Executor, \
    Celery_Step_Executor, Interactive_Step_Executor, analyze_section
from .utils import env, Error, WorkflowDict, get_traceback, ProgressBar, frozendict, dict_merge, short_repr
from .sos_eval import SoS_exec
from .sos_script import SoS_Script
from .sos_syntax import SOS_SECTION_HEADER, SOS_KEYWORDS
from .dag import SoS_DAG
from .target import BaseTarget, FileTarget, UnknownTarget, RemovedTarget, UnavailableLock, sos_variable, textMD5
from .pattern import extract_pattern

__all__ = []


class ExecuteError(Error):
    """Raised when there are errors in prepare mode. Such errors are not raised
    immediately, but will be collected and raised at the end """

    def __init__(self, workflow):
        Error.__init__(self, 'Failed to execute workflow %s' % workflow)
        self.workflow = workflow
        self.errors = []
        self.traces = []
        self.args = (workflow, )

    def append(self, line, error):
        lines = [x for x in line.split('\n') if x.strip()]
        if not lines:
            short_line = '<empty>'
        else:
            short_line = lines[0][:40] if len(lines[0]) > 40 else lines[0]
        if short_line in self.errors:
            return
        self.errors.append(short_line)
        self.traces.append(get_traceback())
        if isinstance(error, Exception):
            self.message += '\n[%s] %s:\n\t%s' % (short_line, error.__class__.__name__, error)
        else:
            self.message += '\n[%s]:\n\t%s' % (short_line, error)

def __null_func__(*args, **kwargs):
    '''This function will be passed to SoS's namespace and be executed
    to evaluate functions of input, output, and depends directives.'''
    return args, kwargs

class Base_Executor:
    '''This is the base class of all executor that provides common
    set up and tear functions for all executors.'''
    def __init__(self, workflow=None, args=[], config_file=None, nested=False):
        env.__task_engine__ = None
        self.workflow = workflow
        self.args = args
        self.nested = nested
        self.config_file = config_file
        # interactive mode does not pass workflow
        if self.workflow:
            self.md5 = self.create_signature()

    def create_signature(self):
        with StringIO() as sig:
            sig.write('# Sections\n')
            for step in self.workflow.sections + self.workflow.auxiliary_sections:
                sig.write('{}: {}\n'.format(step.step_name(), step.md5))
            sig.write('# Command line options\n')
            sig.write('{}\n'.format(self.args))
            self.sig_content = sig.getvalue()
        return textMD5(self.sig_content)[:16]

    def reset_dict(self):
        # if creating a new dictionary, set it up with some basic varibles
        # and functions
        if self.nested:
            SoS_exec('import os, sys, glob')
            SoS_exec('from pysos.runtime import *')
            self._base_symbols = set(dir(__builtins__)) | set(env.sos_dict.keys()) | set(SOS_KEYWORDS) | set(keyword.kwlist)
            self._base_symbols -= {'dynamic'}
            return

        env.sos_dict = WorkflowDict()

        # inject a few things
        env.sos_dict.set('__null_func__', __null_func__)
        env.sos_dict.set('__args__', self.args)
        env.sos_dict.set('__unknown_args__', self.args)
        # initial values
        env.sos_dict.set('SOS_VERSION', __version__)
        env.sos_dict.set('__step_output__', [])

        # load configuration files
        cfg = {}
        sos_config_file = os.path.join(os.path.expanduser('~'), '.sos', 'config.yaml')
        if os.path.isfile(sos_config_file):
            try:
                with open(sos_config_file) as config:
                    cfg = yaml.safe_load(config)
            except Exception as e:
                raise RuntimeError('Failed to parse global sos config file {}, is it in YAML/JSON format? ({})'.format(sos_config_file, e))
        # local config file
        sos_config_file = 'config.yaml'
        if os.path.isfile(sos_config_file):
            try:
                with open(sos_config_file) as config:
                    dict_merge(cfg, yaml.safe_load(config))
            except Exception as e:
                raise RuntimeError('Failed to parse local sos config file {}, is it in YAML/JSON format? ({})'.format(sos_config_file, e))
        # user-specified configuration file.
        if self.config_file is not None:
            if not os.path.isfile(self.config_file):
                raise RuntimeError('Config file {} not found'.format(self.config_file))
            try:
                with open(self.config_file) as config:
                    dict_merge(cfg, yaml.safe_load(config))
            except Exception as e:
                raise RuntimeError('Failed to parse config file {}, is it in YAML/JSON format? ({})'.format(self.config_file, e))
        # set config to CONFIG
        env.sos_dict.set('CONFIG', frozendict(cfg))

        SoS_exec('import os, sys, glob')
        SoS_exec('from pysos.runtime import *')
        self._base_symbols = set(dir(builtins)) | set(env.sos_dict.keys()) | set(SOS_KEYWORDS) | set(keyword.kwlist)
        self._base_symbols -= {'dynamic'}

    def skip(self, section):
        if section.global_def:
            try:
                SoS_exec(section.global_def)
            except RuntimeError as e:
                if env.verbosity > 2:
                    sys.stderr.write(get_traceback())
                raise RuntimeError('Failed to execute statements\n"{}"\n{}'.format(
                    section.global_def, e))
        #
        if 'skip' in section.options:
            val_skip = section.options['skip']
            if val_skip is None or val_skip is True:
                env.logger.info('Step ``{}`` is ``ignored`` due to skip option.'.format(section.step_name()))
                return True
            elif val_skip is not False:
                raise RuntimeError('The value of section option skip can only be None, True or False, {} provided'.format(val_skip))
        return False

    def match(self, target, patterns):
        if isinstance(patterns, (str, BaseTarget)):
            patterns = [patterns]
        elif not isinstance(patterns, Sequence):
            raise RuntimeError('Unknown target to match: {}'.format(patterns))
        #
        for p in patterns:
            # other targets has to match exactly
            if isinstance(target, BaseTarget) or isinstance(p, BaseTarget):
                if target == p:
                    return {}
                else:
                    continue
            # if this is a regular string
            res = extract_pattern(p, [target])
            if res and not any(None in x for x in res.values()):
                return res
            # string match
            elif FileTarget(p) == FileTarget(target):
                return True
        return False

    def resolve_dangling_targets(self, dag, targets=None):
        '''Feed dangling targets with their dependncies from auxiliary steps,
        optionally add other targets'''
        resolved = 0
        while True:
            dangling_targets = dag.dangling(targets)
            if not dangling_targets:
                # if no dangling targets, means all objects COULD be produved by DAG
                break
            env.logger.info('Resolving {} objects from {} nodes'.format(len(dangling_targets), dag.number_of_nodes()))
            # find matching steps
            # check auxiliary steps and see if any steps provides it
            for target in dangling_targets:
                # target might no longer be dangling after a section is added.
                if target not in dag.dangling(targets):
                    continue
                mo = [(x, self.match(target, x.options['provides'])) for x in self.workflow.auxiliary_sections]
                mo = [x for x in mo if x[1] is not False]
                if not mo:
                    for x in self.workflow.auxiliary_sections:
                        env.logger.debug('{}: {}'.format(x.step_name(), x.options['provides']))
                    raise RuntimeError('No step to generate target {}{}'.format(target, dag.steps_depending_on(target, self.workflow)))
                if len(mo) > 1:
                    raise RuntimeError('Multiple steps {} to generate target {}'.format(', '.join(x[0].step_name() for x in mo), target))
                #
                # only one step, we need to process it # execute section with specified input
                #
                # NOTE:  Auxiliary can be called with different output files and matching pattern
                # so we are actually creating a new section each time we need an auxillary step.
                #
                section = mo[0][0]
                if isinstance(mo[0][1], dict):
                    for k,v in mo[0][1].items():
                        env.sos_dict.set(k, v[0])
                #
                # for auxiliary, we need to set input and output, here
                # now, if the step does not provide any alternative (e.g. no variable generated
                # from patten), we should specify all output as output of step. Otherwise the
                # step will be created for multiple outputs. issue #243
                if mo[0][1]:
                    env.sos_dict['__default_output__'] = [target]
                elif isinstance(section.options['provides'], Sequence):
                    env.sos_dict['__default_output__'] = section.options['provides']
                else:
                    env.sos_dict['__default_output__'] = [section.options['provides']]
                # will become input, set to None
                env.sos_dict['__step_output__'] = None
                #
                res = analyze_section(section)
                #
                # build DAG with input and output files of step
                env.logger.info('Adding step {} with output {}'.format(res['step_name'], target))
                if isinstance(mo[0][1], dict):
                    context = mo[0][1]
                else:
                    context = {}
                context['__signature_vars__'] = res['signature_vars']
                context['__environ_vars__'] = res['environ_vars']
                context['__changed_vars__'] = res['changed_vars']
                context['__default_output__'] = env.sos_dict['__default_output__']
                # NOTE: If a step is called multiple times with different targets, it is much better
                # to use different names because pydotplus can be very slow in handling graphs with nodes
                # with identical names.
                dag.add_step(section.uuid, '{} {}'.format(section.step_name(),
                    short_repr(env.sos_dict['__default_output__'])), None, res['step_input'],
                    res['step_depends'], res['step_output'], context=context)
                resolved += 1
        return resolved


    def initialize_dag(self, targets=None):
        '''Create a DAG by analyzing sections statically.'''
        '''Run the script in prepare mode to prepare resources.'''
        # this is for testing only and allows tester to call initialize_dag
        # directly to get a DAG
        if not hasattr(self, '_base_symbols'):
            self.reset_dict()

        dag = SoS_DAG()
        default_input = []
        for idx, section in enumerate(self.workflow.sections):
            if self.skip(section):
                continue
            #
            res = analyze_section(section, default_input)

            environ_vars = res['environ_vars'] - self._base_symbols
            signature_vars = res['signature_vars'] - self._base_symbols
            changed_vars = res['changed_vars']
            # parameters, if used in the step, should be considered environmental
            environ_vars |= env.parameter_vars & signature_vars

            # add shared to targets
            if res['changed_vars']:
                if 'provides' in section.options:
                    if isinstance(section.options['provides'], str):
                        section.options.set('provides', [section.options['provides']])
                else:
                    section.options.set('provides', [])
                #
                section.options.set('provides',
                    section.options['provides'] + [sos_variable(var) for var in changed_vars])

            context={'__signature_vars__': signature_vars,
                    '__environ_vars__': environ_vars,
                    '__changed_vars__': changed_vars }

            # for nested workflow, the input is specified by sos_run, not None.
            if self.nested and idx == 0:
                context['__step_output__'] = env.sos_dict['__step_output__']
            # for regular workflow, the output of the last step has
            # to exist (existence of signature does not count)
            if not self.nested and idx + 1 == len(self.workflow.sections):
                context['__hard_target__'] = True

            # NOTE: if a section has option 'shared', the execution of this step would
            # change dictionary, essentially making all later steps rely on this step.
            dag.add_step(section.uuid,
                section.step_name(),
                idx,
                res['step_input'],
                res['step_depends'],
                res['step_output'],
                context = context)
            default_input = res['step_output']
        #
        # analyze auxiliary steps
        for idx, section in enumerate(self.workflow.auxiliary_sections):
            res = analyze_section(section, default_input)
            environ_vars = res['environ_vars'] - self._base_symbols
            signature_vars = res['signature_vars'] - self._base_symbols
            changed_vars = res['changed_vars']
            # parameters, if used in the step, should be considered environmental
            environ_vars |= env.parameter_vars & signature_vars

            # add shared to targets
            if res['changed_vars']:
                if 'provides' in section.options:
                    if isinstance(section.options['provides'], str):
                        section.options.set('provides', [section.options['provides']])
                else:
                    section.options.set('provides', [])
                #
                section.options.set('provides',
                    section.options['provides'] + [sos_variable(var) for var in changed_vars])
        #
        self.resolve_dangling_targets(dag, targets)
        # now, there should be no dangling targets, let us connect nodes
        dag.build(self.workflow.auxiliary_sections)
        # trim the DAG if targets are specified
        if targets:
            dag = dag.subgraph_from(targets)
        # write DAG for debugging purposes
        dag.write_dot(os.path.join(env.exec_dir, '.sos', '{}.dot'.format(self.workflow.name)))
        # check error
        cycle = dag.circular_dependencies()
        if cycle:
            raise RuntimeError('Circular dependency detected {}. It is likely a later step produces input of a previous step.'.format(cycle))

        return dag

    def save_workflow_signature(self, dag):
        '''Save tracked files in .sos so that untracked files can be cleaned by command
        sos clean.
        '''
        with open(os.path.join(env.exec_dir, '.sos', '{}.sig'.format(self.md5)), 'w') as sigfile:
            sigfile.write(self.sig_content)
            sigfile.write('# input and dependent files\n')
            for target in sorted(x for x in dag._all_dependent_files if isinstance(x, str)):
                sigfile.write('{}\n'.format(target))
            sigfile.write('# output files\n')
            for target in sorted(x for x in dag._all_output_files if isinstance(x, str)):
                sigfile.write('{}\n'.format(target))   
        
    def run(self, targets=None, mode='run'):
        '''Execute a workflow with specified command line args. If sub is True, this
        workflow is a nested workflow and be treated slightly differently.
        '''
        self.reset_dict()
        env.run_mode = mode
        # passing run_mode to SoS dict so that users can execute blocks of
        # python statements in different run modes.
        env.sos_dict.set('run_mode', env.run_mode)
        # process step of the pipelinp
        dag = self.initialize_dag(targets=targets)
        #
        # if targets are specified and there are only signatures for them, we need
        # to remove the signature and really generate them
        if targets:
            for t in targets:
                if not FileTarget(t).exists('target'):
                    FileTarget(t).remove('signature')
        #
        SoS_exec('from pysos.runtime import sos_handle_parameter_')
        #
        prog = ProgressBar(self.workflow.name, dag.num_nodes(), disp=dag.num_nodes() > 1 and env.verbosity == 1)
        self.reset_dict()
        env.sos_dict.set('run_mode', env.run_mode)
        exec_error = ExecuteError(self.workflow.name)
        while True:
            # find any step that can be executed and run it, and update the DAT
            # with status.
            runnable = dag.find_executable()
            if runnable is None:
                # no runnable
                #dag.show_nodes()
                break
            # find the section from runnable
            section = self.workflow.section_by_id(runnable._step_uuid)
            #
            # this is to keep compatibility of dag run with sequential run because
            # in sequential run, we evaluate global section of each step in
            # order to determine values of options such as skip.
            # The consequence is that global definitions are available in
            # SoS namespace.
            try:
                SoS_exec(section.global_def)
            except Exception as e:
                if env.verbosity > 2:
                    sys.stderr.write(get_traceback())
                raise RuntimeError('Failed to execute statements\n"{}"\n{}'.format(
                    section.global_def, e))

            # clear existing keys, otherwise the results from some random result
            # might mess with the execution of another step that does not define input
            for k in ['__step_input__', '__default_output__', '__step_output__']:
                if k in env.sos_dict:
                    env.sos_dict.pop(k)
            # if the step has its own context
            env.sos_dict.quick_update(runnable._context)
            # execute section with specified input
            runnable._status = 'running'
            q = mp.Queue()
            if mode == 'run':
                executor = SP_Step_Executor(section, q)
            elif mode == 'prepare':
                executor = Prepare_Step_Executor(section, q)
            else:
                executor = Dryrun_Step_Executor(section, q)
            p = mp.Process(target=executor.run)
            p.start()
            #
            res = q.get()
            # if we does get the result
            p.join()
            # if the step says unknown target .... need to check if the target can
            # be build dynamically.
            if isinstance(res, UnknownTarget):
                runnable._status = None
                target = res.target
                if self.resolve_dangling_targets(dag, [target]) == 0:
                    raise RuntimeError('Failed to resolve {}{}.'
                        .format(target, dag.steps_depending_on(target, self.workflow)))
                # now, there should be no dangling targets, let us connect nodes
                # this can be done more efficiently
                dag.build(self.workflow.auxiliary_sections)
                cycle = dag.circular_dependencies()
                if cycle:
                    raise RuntimeError('Circular dependency detected {}. It is likely a later step produces input of a previous step.'.format(cycle))
            elif isinstance(res, RemovedTarget):
                runnable._status = None
                dag.regenerate_target(res.target)
            elif isinstance(res, UnavailableLock):
                runnable._status = 'pending'
                runnable._signature = (res.output, res.sig_file)
                env.logger.info('Waiting on another process for step {}'.format(section.step_name()))
            # if the job is failed
            elif isinstance(res, Exception):
                runnable._status = 'failed'
                exec_error.append(runnable._node_id, res)
                prog.progress(1)
            else:#
                for k, v in res.items():
                    env.sos_dict.set(k, v)
                #
                # set context to the next logic step.
                for edge in dag.out_edges(runnable):
                    node = edge[1]
                    # if node is the logical next step...
                    if node._node_index is not None and runnable._node_index is not None:
                        #and node._node_index == runnable._node_index + 1:
                        node._context.update(env.sos_dict.clone_selected_vars(
                            node._context['__signature_vars__'] | node._context['__environ_vars__'] \
                            | {'_input', '__step_output__', '__default_output__', '__args__'}))
                runnable._status = 'completed'
                prog.progress(1)
            #env.logger.error('completed')
        prog.done()
        if exec_error.errors:
            failed_steps, pending_steps = dag.pending()
            if failed_steps:
                sections = [self.workflow.section_by_id(x._step_uuid).step_name() for x in failed_steps]
                exec_error.append(self.workflow.name,
                    RuntimeError('{} failed step{}: {}'.format(len(sections), 
                        's' if len(sections) > 1 else '', ', '.join(sections))))
            if pending_steps:
                sections = [self.workflow.section_by_id(x._step_uuid).step_name() for x in pending_steps]
                exec_error.append(self.workflow.name,
                    RuntimeError('{} pending step{}: {}'.format(len(sections),
                        's' if len(sections) > 1 else '', ', '.join(sections))))
            raise exec_error
        else:
            self.save_workflow_signature(dag)
            env.logger.info('Workflow {} (ID={}) is executed successfully.'.format(self.workflow.name, self.md5))

    def dryrun(self, targets=None):
        '''Execute the script in dryrun mode.'''
        try:
            self.run(targets=targets, mode='dryrun')
        # only runtime errors are ignored
        except RuntimeError as e:
            env.logger.warning('Workflow cannot be completed in dryrun mode: {}'.format(e))

    def prepare(self, targets=None):
        '''Execute the script in prepare mode.'''
        try:
            self.run(targets=targets, mode='prepare')
        # only runtime errors are ignored
        except RuntimeError as e:
            env.logger.warning('Workflow cannot be completed in prepare mode: {}'.format(e))

class MP_Executor(Base_Executor):
    #
    # Execute a workflow sequentially in batch mode
    def __init__(self, workflow, args=[], config_file=None, nested=False):
        Base_Executor.__init__(self, workflow, args, config_file, nested=nested)
        if hasattr(env, 'accessed_vars'):
            delattr(env, 'accessed_vars')

    def step_executor(self, section, queue):
        return MP_Step_Executor(section, queue)

    def run(self, targets=None, mode='run'):
        '''Execute a workflow with specified command line args. If sub is True, this
        workflow is a nested workflow and be treated slightly differently.
        '''
        self.reset_dict()
        env.run_mode = mode
        # passing run_mode to SoS dict so that users can execute blocks of
        # python statements in different run modes.
        env.sos_dict.set('run_mode', env.run_mode)
        # process step of the pipelinp
        dag = self.initialize_dag(targets=targets)
        #
        SoS_exec('from pysos.runtime import sos_handle_parameter_')

        # process step of the pipelinp
        #
        procs = [None for x in range(env.max_jobs)]
        prog = ProgressBar(self.workflow.name, dag.num_nodes(), disp=dag.num_nodes() > 1 and env.verbosity == 1)
        exec_error = ExecuteError(self.workflow.name)
        while True:
            # step 1: check existing jobs and see if they are completed
            for idx, proc in enumerate(procs):
                if proc is None:
                    continue
                (p, q, u) = proc
                try:
                    res = q.get_nowait()
                except Empty:
                    # continue waiting
                    continue
                #
                # if we does get the result
                p.join()

                runnable = dag.node_by_id(u)
                if isinstance(res, UnknownTarget):
                    runnable._status = None
                    target = res.target
                    if self.resolve_dangling_targets(dag, [target]) == 0:
                        raise RuntimeError('Failed to resolve {}{}.'
                            .format(target, dag.steps_depending_on(target, self.workflow)))
                    # now, there should be no dangling targets, let us connect nodes
                    # this can be done more efficiently
                    dag.build(self.workflow.auxiliary_sections)
                    cycle = dag.circular_dependencies()
                    if cycle:
                        raise RuntimeError('Circular dependency detected {}. It is likely a later step produces input of a previous step.'.format(cycle))
                elif isinstance(res, RemovedTarget):
                    runnable._status = None
                    dag.regenerate_target(res.target)
                elif isinstance(res, UnavailableLock):
                    runnable._status = 'pending'
                    runnable._signature = (res.output, res.sig_file)
                    section = self.workflow.section_by_id(runnable._step_uuid)
                    env.logger.info('Waiting on another process for step {}'.format(section.step_name()))
                    # move away to let other tasks to run first
                    procs[idx] = None
                # if the job is failed
                elif isinstance(res, Exception):
                    runnable._status = 'failed'
                    exec_error.append(runnable._node_id, res)
                    prog.progress(1)
                    procs[idx] = None
                else:
                    #
                    for k, v in res.items():
                        env.sos_dict.set(k, v)
                    #
                    # set context to the next logic step.
                    for edge in dag.out_edges(runnable):
                        node = edge[1]
                        # if node is the logical next step...
                        if node._node_index is not None and runnable._node_index is not None:
                            #and node._node_index == runnable._node_index + 1:
                            node._context.update(env.sos_dict.clone_selected_vars(
                                node._context['__signature_vars__'] | node._context['__environ_vars__'] \
                                | {'_input', '__step_output__', '__default_output__', '__args__'}))
                    runnable._status = 'completed'
                    prog.progress(1)
                    procs[idx] = None
                #env.logger.error('completed')
                #dag.show_nodes()
            # step 2: submit new jobs if there are empty slots
            for idx, proc in enumerate(procs):
                # if there is empty slot, submit
                if proc is not None:
                    continue
                # find any step that can be executed and run it, and update the DAT
                # with status.
                runnable = dag.find_executable()
                if runnable is None:
                    # no runnable
                    #dag.show_nodes()
                    break
                # find the section from runnable
                section = self.workflow.section_by_id(runnable._step_uuid)
                #
                # this is to keep compatibility of dag run with sequential run because
                # in sequential run, we evaluate global section of each step in
                # order to determine values of options such as skip.
                # The consequence is that global definitions are available in
                # SoS namespace.
                try:
                    SoS_exec(section.global_def)
                except RuntimeError as e:
                    if env.verbosity > 2:
                        sys.stderr.write(get_traceback())
                    raise RuntimeError('Failed to execute statements\n"{}"\n{}'.format(
                        section.global_def, e))

                # clear existing keys, otherwise the results from some random result
                # might mess with the execution of another step that does not define input
                for k in ['__step_input__', '__default_output__', '__step_output__']:
                    if k in env.sos_dict:
                        env.sos_dict.pop(k)
                # if the step has its own context
                env.sos_dict.quick_update(runnable._context)
                # execute section with specified input
                runnable._status = 'running'
                q = mp.Queue()
                executor = self.step_executor(section, q)
                p = mp.Process(target=executor.run)
                procs[idx] = (p, q, runnable._node_uuid)
                p.start()
                #
                #env.logger.error('started')
                #dag.show_nodes()
            #
            if all(x is None for x in procs):
                break
            else:
                time.sleep(0.1)
        prog.done()
        if exec_error.errors:
            failed_steps, pending_steps = dag.pending()
            if failed_steps:
                sections = [self.workflow.section_by_id(x._step_uuid).step_name() for x in failed_steps]
                exec_error.append(self.workflow.name,
                    RuntimeError('{} failed step{}: {}'.format(len(sections), 
                        's' if len(sections) > 1 else '', ', '.join(sections))))
            if pending_steps:
                sections = [self.workflow.section_by_id(x._step_uuid).step_name() for x in pending_steps]
                exec_error.append(self.workflow.name,
                    RuntimeError('{} pending step{}: {}'.format(len(sections), 
                        's' if len(sections) > 1 else '', ', '.join(sections))))
            raise exec_error
        else:
            sig = self.save_workflow_signature(dag)
            env.logger.info('Workflow {} (ID={}) is executed successfully.'.format(self.workflow.name, sig))

class RQ_Executor(MP_Executor):
    def __init__(self, workflow, args=[], config_file=None, nested=False):
        MP_Executor.__init__(self, workflow, args, config_file, nested=nested)
        env.__task_engine__ = 'rq'

        from rq import Queue as rqQueue
        from redis import Redis
        import yaml

        connection_file = os.path.join(env.exec_dir, '.sos', 'redis_connection.yaml')
        if os.path.isfile(connection_file):
            try:
                with open(connection_file) as conn:
                    cfg = yaml.safe_load(conn)
                if cfg is None:
                    cfg = {}
            except Exception as e:
                env.logger.error('Failed to parse redis connection file {}. ({}}'.format(connection_file, e))
                sys.exit(1)
            try:
                redis_conn = Redis(host=cfg['host'], port=cfg.get('port', 6379))
            except Exception as e:
                env.logger.error('Failed to connect to redis server: {}'.format(e))
                sys.exit(1)
        else:
            redis_conn = Redis()
        self.redis_queue = rqQueue(connection=redis_conn)

    def step_executor(self, section, queue):
        return RQ_Step_Executor(section, queue, self.redis_queue)

class Celery_Executor(MP_Executor):
    def __init__(self, workflow, args=[], config_file=None, nested=False):
        MP_Executor.__init__(self, workflow, args, config_file, nested=nested)
        env.__task_engine__ = 'Celery'

    def step_executor(self, section, queue):
        # pass celery_app if needed
        return Celery_Step_Executor(section, queue)

class Interactive_Executor(Base_Executor):
    '''Interactive executor called from by iPython Jupyter or Spyder'''
    def __init__(self):
        # we actually do not have our own workflow, everything is passed from ipython
        # by nested = True we actually mean no new dictionary
        Base_Executor.__init__(self, nested=True)

    def parse_command_line(self, command_line):
        parser = argparse.ArgumentParser()
        # no default workflow so it will execute any workflow if the code piece
        # defines only one workflow
        # 
        # parser.add_argument('-j', type=int, metavar='JOBS', default=1, dest='__max_jobs__')
        parser.add_argument('-c', dest='__config__', metavar='CONFIG_FILE')
        #parser.add_argument('-r', dest='__report__', metavar='REPORT_FILE',
        #    default=os.path.join('.sos', '__step_report.md'))
        runmode = parser.add_argument_group(title='Run mode options')
        runmode.add_argument('-f', action='store_true', dest='__rerun__')
        runmode.add_argument('-F', action='store_true', dest='__construct__')
        # default to 1 to avoid output env.logger.info to notebook
        parser.add_argument('-v', '--verbosity', type=int, choices=range(5), default=1)
        #
        args, workflow_args = parser.parse_known_args(shlex.split(command_line))
        return args, workflow_args

    def parse_script(self, code):
        '''Used by the kernel to judge if the code is complete'''
        return SoS_Script(content=code)

    def set_dict(self, args):
        env.sos_dict.set('__null_func__', __null_func__)
        env.sos_dict.set('SOS_VERSION', __version__)

        # load configuration files
        cfg = {}
        sos_config_file = os.path.join(os.path.expanduser('~'), '.sos', 'config.yaml')
        if os.path.isfile(sos_config_file):
            try:
                with open(sos_config_file) as config:
                    cfg = yaml.safe_load(config)
            except Exception as e:
                raise RuntimeError('Failed to parse global sos config file {}, is it in YAML/JSON format? ({})'.format(sos_config_file, e))
        # local config file
        sos_config_file = 'config.yaml'
        if os.path.isfile(sos_config_file):
            try:
                with open(sos_config_file) as config:
                    dict_merge(cfg, yaml.safe_load(config))
            except Exception as e:
                raise RuntimeError('Failed to parse local sos config file {}, is it in YAML/JSON format? ({})'.format(sos_config_file, e))
        if args.__config__ is not None:
            # user-specified configuration file.
            if not os.path.isfile(args.__config__):
                raise RuntimeError('Config file {} not found'.format(args.__config__))
            try:
                with open(args.__config__) as config:
                    dict_merge(cfg, yaml.safe_load(config))
            except Exception as e:
                raise RuntimeError('Failed to parse config file {}, is it in YAML/JSON format? ({})'.format(self.config_file, e))
        # set config to CONFIG
        env.sos_dict.set('CONFIG', frozendict(cfg))

    def run(self, block, command_line=''):
        '''Execute a block of SoS script that is sent by iPython/Jupyer/Spyer
        The code can be simple SoS/Python statements, one SoS step, or more
        or more SoS workflows with multiple steps. This executor,
        1. adds a section header to the script if there is no section head
        2. execute the workflow in interactive mode, which is different from
           batch mode in a number of ways, which most notably without support
           for nested workflow.
        3. Optionally execute the workflow in preparation mode for debugging purposes.
        '''
        # if there is no valid code do nothing
        if not block.strip():
            return
        # if there is no section header, add a header so that the block
        # appears to be a SoS script with one section
        if not any([SOS_SECTION_HEADER.match(line) for line in block.split()]):
            block = '[interactive_0]\n' + block

        script = SoS_Script(content=block)
        env.run_mode = 'interactive'
        try:
            args, workflow_args = self.parse_command_line(command_line)
            env.sos_dict.set('__args__', workflow_args)
            env.sos_dict.set('__unknown_args__', workflow_args)
            self.set_dict(args)
            self.workflow = script.workflow()

            if args.__rerun__:
                env.sig_mode = 'ignore'
            elif args.__construct__:
                env.sig_mode = 'construct'
            else:
                env.sig_mode = 'default'

            #if os.path.isfile(args.__report__):
            #    os.remove(args.__report__)

            # this is the result returned by the workflow, if the
            # last stement is an expression.
            last_res = None
            #
            # clear __step_input__, __step_output__ etc because there is
            # no concept of passing input/outputs across cells.
            env.sos_dict.set('__step_output__', [])
            for k in ['__step_input__', '__default_output__', 'input', 'output', \
                'depends', '_input', '_output', '_depends']:
                env.sos_dict.pop(k, None)

            for idx, section in enumerate(self.workflow.sections):
                if 'skip' in section.options:
                    val_skip = section.options['skip']
                    if val_skip is None or val_skip is True:
                        continue
                    elif val_skip is not False:
                        raise RuntimeError('The value of section option skip can only be None, True or False, {} provided'.format(val_skip))
                #
                last_res = Interactive_Step_Executor(section).run()
                # if the step is failed
                if isinstance(last_res, Exception):
                    raise RuntimeError(last_res)
            return last_res
        finally:
            env.verbosity = 1
            env.sig_mode = 'default'



